import time
import warnings
import numpy as np
from astropy.io import fits as pyfits
from astropy.time import Time
import scipy.interpolate as spi
from scipy.spatial.qhull import QhullError
from dkist_processing_pac import Data, generic
from dkist_processing_pac.tag import tag

warnings.simplefilter('always', UserWarning)
ATOL = 1e-6
RTOL = 1e-6


class TelescopeModel():
    """Build up the Mueller matrix of the full "Telescope Model" for use in PA&C analysis. As detailed in the DKIST
    PolCal Plan this model is parametrized by 3 mirror groups (M12, M34, M56) and the rotation matrices between them.
    The mirror Mueller matrices are calculated in real time from the parameters x (the ration of reflectivity between
    light parallel and perpendicular to the plane of incidence) and \tau (the retardance). The rotation matrices are
    also calculated in real time based on the (alt, az, coude_table) angles of DKIST.

    Each of the matrices in the Telescope Model are individually accessible as properties of the class
    (e.g. TelescopeModel.M34) and the full model exists in the .TM property. Note that the Telescope Model does NOT
    include the M12 group, but that group's Mueller matrix is still included in this object.

    Because each component of the model is recomputed each time it is queried this class lends itself well to iterative
    fitting.

    Attributes
    (only those you might need)
    ----------
    TM : numpy.ndarray
        Set of Mueller matrices for the entire telescope model. Shape (N, 4, 4)

    M12 : numpy.ndarray
        Mueller matrix of the M12 group. Shape (4, 4)

    R23 : numpy.ndarray
        Mueller matrix for rotation between M2 and M3. Shape (4, 4)

    M34 : numpy.ndarray
        Mueller matrix of the M34 group. Shape (4, 4)

    R45 : numpy.ndarray
        Mueller matrix for rotation between M4 and M5. Shape (N, 4, 4)

    M56 : numpy.ndarray
        Mueller matrix of the M56 group. Shape (4, 4)

    numsteps : int
        Number of steps in a Calibration Sequence (N)
    """

    def __init__(self, azimuth, elevation, table_angle):
        """Initialize the class with a sequence of telescope geometries.

        If the inputs are array-like then each entry is assumed to correspond to an individual step in the calibration
        sequence.

        Parameters
        ----------
        azimuth : float or array-like
            Telescope azimuth angle (in degrees)

        elevation : float or array-like
            Telescope elevation angle (in degrees)

        table_angle : float or array-like
            Angle (in degrees) of coude table
        """
        self.x12 = 1.07
        self.t12 = np.pi
        self.x34 = 1.07
        self.t34 = np.pi
        self.x56 = 1.07
        self.t56 = np.pi
        self.elevation = np.atleast_1d(elevation) * np.pi / 180
        self.azimuth = np.atleast_1d(azimuth) * np.pi / 180
        self.table_angle = np.atleast_1d(table_angle) * np.pi / 180

        if self.elevation.shape != self.azimuth.shape or self.elevation.shape != self.table_angle.shape:
            raise ValueError('Telescope geometry vectors do not have the same shape')

        self.numsteps = self.elevation.size

        self.wavelength = None
        self.obstime = None

    def __add__(self, other):
        """Concatenate two Telescope Models together.

        Returns
        -------
        TelescopeModel
            A new TelescopeModel object
        """
        output = TelescopeModel(np.append(self.azimuth, other.azimuth) * 180 / np.pi,
                                np.append(self.elevation, other.elevation) * 180 / np.pi,
                                np.append(self.table_angle, other.table_angle) * 180 / np.pi)

        return output

    def load_from_database(self, db_file, obstime, obswave, method='linear'):
        """Update (x, t) mirror parameters based on a database of previous measurements.

        Given a date and wavelength, the closest set of parameters is found via interpolation. The default is to use
        linear interpolation, but cubic interpolation can be requested via the `method` parameter.

        If the supplied time or wavelength is outside of the parameter space covered by the database then the values
        are set to the closest (time, wave) coordinate rather than extrapolating.

        Parameters
        ----------
        db_file : str
            The path to the database file (see Notes)

        obstime : float
            The time at which to interpolate the parameters. Format is MJD.

        obswave : float
            The wavelength at which to interpolate the parameters (nm)

        method : str
            The interpolation method to use. Can be either 'nearest', 'linear' (default), or 'cubic'

        Notes
        -----
        Currently the database is simply a space-delimited text file with the following columns:

            MJD wavelength x12 t12 x34 t34 x56 t56

        """
        times, wave, x12, t12, x34, t34, x56, t56 = np.loadtxt(db_file, unpack=True)
        print(f"{tag()}: Loading database parameters from {db_file} for {obswave} at {Time(obstime, format='mjd').fits}")

        for source, target in zip([x12, t12, x34, t34, x56, t56],
                                  ['x12', 't12', 'x34', 't34', 'x56', 't56']):

            try:
                value = float(spi.griddata((times, wave), source, (obstime, obswave), method=method))
            except QhullError:
                warnings.warn('Telescope database lacks diversity along either time or wavelength axis. '
                              'Reverting to nearest neighbor interpolation.')
                value = float(spi.griddata((times, wave), source, (obstime, obswave), method='nearest'))


            # griddata returns NaN if out of range
            if np.isnan(value):
                warnings.warn(
                        'Requested time/wavelength is outside of the Telescope Database range, using the nearest '
                        'in-bounds value')
                value = float(spi.griddata((times, wave), source, (obstime, obswave), method='nearest'))

            setattr(self, target, value)
            
        self.wavelength = obswave
        self.obstime = obstime
        print('{:}: loaded (x12, t12, x34, t34, x56, t56) = ({:.5f}, {:.5f}, {:.5f}, {:.5f}, {:.5f}, {:.5f}) at {:} nm'.format(tag(),
                                                                                                                               self.x12,
                                                                                                                               self.t12,
                                                                                                                               self.x34,
                                                                                                                               self.t34,
                                                                                                                               self.x56,
                                                                                                                               self.t56,
                                                                                                                               self.wavelength))

    def save_to_database(self, db_file, obstime=None, obswave=None):
        """Update the telescope database with the current class telescope mirror parameters.

        If the wavelength and time are not set in the class they must be passed explicitly.

        Parameters
        ----------
        db_file : str
            Path to parameter database. See notes for load_from_database for information on the format of this file.

        obstime : float
            The time of the current observation, in MJD. Can also be 'now' to use the current system time.

        obswave : float
            The wavelength (nm) of the current observation.
        """
        if obstime == 'now':
            new_time = Time(time.time(), format='unix').mjd
        elif obstime is None:
            if self.obstime is None:
                raise ValueError('Obstime was not set in the Telescope Model nor specified explicitly')
            else:
                new_time = self.obstime
        else:
            new_time = obstime

        if obswave is None:
            if self.wavelength is None:
                raise ValueError('Wavelength was not set in the Telescope Model nor specified explicitly')
            else:
                new_wave = self.wavelength
        else:
            new_wave = obswave

        with open(db_file, 'a') as f:
            f.write(str('{:15.6f}' + '{:8.2f}' + '{:7.3f}' * 6 + '\n').format(new_time,
                                                                              new_wave,
                                                                              self.x12,
                                                                              self.t12,
                                                                              self.x34,
                                                                              self.t34,
                                                                              self.x56,
                                                                              self.t56))

    def save_FITS(self, outputfile, overwrite=False):
        """Save a FITS file with the M3 - M7 Mueller matrix and M12 Muller matrix as separate HDUs.

        It's not immediately obvious why this method is useful. Wouldn't it make more sense to just load the object
        itself and use the associated properties? You're going to need to load the object anyway to enter the correct
        telescope geometry.
        """

        if self.numsteps > 1:
            warnings.warn('Multiple telescope geometries found. Only using the first configuration')

        primary = pyfits.PrimaryHDU()
        primary.header['TAZIMUTH'] = self.azimuth[0]
        primary.header['TELEVATN'] = self.elevation[0]
        primary.header['TTBLANGL'] = self.table_angle[0]
        primary.header['X12'] = self.x12
        primary.header['T12'] = self.t12
        primary.header['X34'] = self.x34
        primary.header['T34'] = self.t34
        primary.header['X56'] = self.x56
        primary.header['T56'] = self.t56

        tm = pyfits.ImageHDU(self.TM[0, :, :])
        tm.header['NAME'] = ('TM', 'Mueller matrix for mirrors 3-6')

        m12 = pyfits.ImageHDU(self.M12)
        m12.header['NAME'] = ('M12', 'Mueller matrix for M12 group')

        pyfits.HDUList([primary, tm, m12]).writeto(outputfile, overwrite=overwrite)

    def generate_inverse_telescope_model(self, M12=True, include_parallactic=True):
        """Produce the inverse of the full Telescope Model's Mueller matrix.

        The user can choose to include M12 as part of the Telescope Model, in which case the inverse will capture all
        polarization effects between the DKIST entrance aperture and M7.

        If, for whatever reason, the generated inverse does not satisfy T int(T) = inv(T) T = Identity then an error
        will be raised.

        Parameters
        ----------
        M12 : bool
            If True then include M12 in the Telescope Model

        include_parallactic : bool
            If True then the final rotation from DKIST to Solar reference frame will be included in the output matrix

        Returns
        -------
        numpy.ndarray
            The (4, 4) inverse Mueller matrix of the Telescope Model.
        """
        full_model = self.TM

        if M12:
            full_model = full_model @ self.M12

        if full_model.shape[0] > 1:
            warnings.warn('Multiple telescope geometries found. Only using the first configuration')

        inverse = np.linalg.inv(full_model[0, :, :])

        if not (np.allclose(np.diag(np.ones(4)), inverse @ full_model[0], rtol=RTOL, atol=ATOL) and
                np.allclose(np.diag(np.ones(4)), full_model[0] @ inverse, rtol=RTOL, atol=ATOL)):
            raise ArithmeticError('The generated inverse is not mathematically valid')

        if include_parallactic:
            time = Time(self.obstime, format='mjd')
            p_rot = Data.rotation(-1 * generic.compute_parallactic_angle(time))
            inverse = p_rot @ inverse

        return inverse

    @property
    def M12(self):
        """The M12 mirror Mueller matrix"""
        return Data.mirror(self.x12, self.t12)

    @property
    def M34(self):
        """The M34 mirror Mueller matrix"""
        return Data.mirror(self.x34, self.t34)

    @property
    def M56(self):
        """The M56 mirror Mueller matrix"""
        return Data.mirror(self.x56, self.t56)

    @property
    def R23(self):
        """The rotation matrix between M2 and M3. This is always the same, so it doesn't have a step dimension

        Returns
        -------
        numpy.ndarray
            Array of shape (4, 4)
        """
        return Data.rotation(-np.pi / 2.)

    @property
    def R45(self):
        """The rotation matrix between M4 and M5

        Returns
        -------
        numpy.ndarray
            Array of shape (N, 4, 4)
        """

        Rarr = np.empty((self.numsteps, 4, 4), dtype=np.float64)
        for i in range(self.numsteps):
            Rarr[i, :, :] = Data.rotation(-1 * self.elevation[i])

        return Rarr

    @property
    def R67(self):
        """The rotation matrix between M6 and M7

        Returns
        -------
        numpy.ndarray
            Array of shape (N, 4, 4)
        """

        Rarr = np.empty((self.numsteps, 4, 4), dtype=np.float64)
        for i in range(self.numsteps):
            theta = self.azimuth[i] - self.table_angle[i] + np.pi / 2.
            Rarr[i, :, :] = Data.rotation(theta)

        return Rarr

    @property
    def TM(self):
        """The completed Telescope Model

        Returns
        -------
        numpy.ndarray
            Array of shape (N, 4, 4)
        """
        return self.R67 @ self.M56 @ self.R45 @ self.M34 @ self.R23



import pkg_resources
import asdf
import numpy as np
from astropy.time import Time

def make_polcal_init_val_asdf(system_file: str,
                              minmax_frac: float = 0.02,
                              valid_time: str = 'now',
                              set_name: str = 'default') -> None:

    sys_wave, ret0h, ret045, ret0r, t_ret, t_pol, I_sys, x12, t12, x34, t34, x56, t56 = np.loadtxt(system_file,
                                                                                                   unpack=True,
                                                                                                   usecols=range(13))

    minmax_mult = np.array([-minmax_frac, 0, minmax_frac]) + 1.
    minmax_mult = minmax_mult[None, :]
    Q_in = (1 - x12**2) / (1 + x12**2)

    if valid_time == 'now':
        valid_time = Time.now().fits

    ## Just a dummy for now. We'll likely never use this, but it's included for backwards compatibility
    dummy_dret = np.zeros(sys_wave.shape) + 0.1 * np.pi / 180.
    dummy_QU = np.zeros((sys_wave.size, 3)) + np.array([-0.1, 0, 0.1])[None, :]
    dummy_Isys = np.zeros((sys_wave.size, 3)) + np.array([-0.1, 1, 0.1])[None, :]

    polcal_tree = {'gen_time': Time.now().fits,
                   'valid_time': valid_time,
                   'wave': sys_wave,
                   'params': {'ret0h': ret0h[:, None] * minmax_mult,
                              'dreth': dummy_dret[:, None] * minmax_mult,
                              'ret045': ret045[:, None] * minmax_mult,
                              'dret45': dummy_dret[:, None] * minmax_mult,
                              'ret0r': ret0r[:, None] * minmax_mult,
                              'dretr': dummy_dret[:, None] * minmax_mult,
                              'I_sys': dummy_Isys,
                              't_ret': t_ret[:, None] * minmax_mult,
                              't_pol': t_pol[:, None] * minmax_mult,
                              'Q_in': Q_in[:, None] * minmax_mult,
                              'U_in': dummy_QU,
                              'V_in': dummy_QU}
                   }

    initval_file = pkg_resources.resource_filename('dkist_processing_pac', 'data/init_values/polcal_{}.asdf'.format(set_name))
    af = asdf.AsdfFile(polcal_tree)
    af.write_to(initval_file)

    return

def make_groupcal_init_val_asdf(system_file: str,
                                minmax_frac: float = 0.02,
                                valid_time: str = 'now',
                                set_name: str = 'default') -> None:

    sys_wave, ret0h, ret045, ret0r, t_ret, t_pol, I_sys, x12, t12, x34, t34, x56, t56 = np.loadtxt(system_file,
                                                                                                   unpack=True,
                                                                                                   usecols=range(13))

    minmax_mult = np.array([-minmax_frac, 0, minmax_frac]) + 1.
    minmax_mult = minmax_mult[None, :]

    if valid_time == 'now':
        valid_time = Time.now().fits

    groupcal_tree = {'gen_time': Time.now().fits,
                     'valid_time': valid_time,
                     'wave': sys_wave,
                     'params': {'x34': x34[:, None] * minmax_mult,
                                't34': t34[:, None] * minmax_mult,
                                'x56': x56[:, None] * minmax_mult,
                                't56': t56[:, None] * minmax_mult}
                     }

    initval_file = pkg_resources.resource_filename('dkist_processing_pac', 'data/init_values/groupcal_{}.asdf'.format(set_name))
    af = asdf.AsdfFile(groupcal_tree)
    af.write_to(initval_file)

    return



import pkg_resources
import numpy as np
from dkist_processing_pac import Data
from dkist_processing_pac.tag import tag


class CalibrationSequence():
    """Set up the Mueller matrices of the Calibration Unit at each step in a Calibration Sequence. The sequence is
    defined by vectors containing the angles of the polarizer and retarder, but each of these elements can either be
    inserted or removed from the beam as necessary.

    The complete model of the Calibration Unit consists of a polarizer followed by an elliptical retarder. The set of
    Mueller matrices for each element can be accessed directly and are updated in real time whenever they are queried.

    The set of N (number of steps) Mueller matrices for the entire CU is accessed via the .CM property.

    The fact that all components of the model are recomputed each time they are queried makes this class a natural fit
    for iterative fitting techniques.

    Multiple Calibration Sequences can be strung together with the + operator. This simply stores all CS's in the same
    object; each CS is allowed to have its own set of parameters.

    Attributes
    (only those you might need)
    ----------
    CM : numpy.ndarray
        Set of Mueller matrices for the entire CU at each step. Shape (N, 4, 4)

    S_in : numpy.ndarray
        The Stokes vector of idea input light. Defaults to [1,0,0,0]

    ret : numpy.ndarray
        The value of retardance computed at each step in the CS

    pol_mat : numpy.ndarray
        Set of Mueller matrices for the polarizer alone. Shape (N, 4, 4)

    ret_mat : numpy.ndarray
        Set of Mueller matrices for the retarder alone. Shape (N, 4, 4)

    Notes
    -----
    The current CU model has six parameters: the intercept and slope for horizontal, 45 degree, and circular retardance.
    Each of the three retardance values is given by

            ret_i(t) = ret_0_i + dret_i * (t - t_0),

    where ret_0_i is the retardance at time t_0.

    """

    def __init__(self, theta_pol_steps, theta_ret_steps, pol_in, ret_in, dark_in, timeobs,
                 ret_0_h=0., dret_h=0.,
                 ret_0_45=0., dret_45=0.,
                 ret_0_r=0., dret_r=0.,
                 t_pol_0=1., t_ret_0=1., py=0,
                 I_sys = 1., Q_in=0, U_in=0, V_in=0):
        """Initialize the Calibration Unit at each of the polarizer/retarder angles provided in the Calibration Sequence

        The values for the properties of the optical elements and the angular uncertainty can also be provided.

        Parameters
        ----------
        theta_pol_steps : numpy.ndarray
            A 1D array of length N that has the angles (in degrees) of the polarizer at each step of the CS

        theta_ret_steps : numpy.ndarray
            A 1D array of length N that has the angles (in degrees) of the retarder at each step of the CS

        pol_in : numpy.ndarray
            A 1D Boolean array of length N that describes the state of the GOS polarizer at each step in the CS.
            True corresponds to the polarizer in the beam path.

        ret_in : numpy.ndarray
            A 1D Boolean array of length N that describes the state of the GOS retarder at each step in the CS.
            True corresponds to the retarder in the beam path.

        timeobs : numpy.ndarray
            A 1D array of length N containing the observation time (MJD) of each step of the CS.

        ret_0_i : float
            Value of the retardance i intercept (in radians).

        dret_i : float
            Value of the retardance i slope (in radians).

        py : float
            Transmission coefficient of the E_y electric field through the polarizer
        """
        if theta_ret_steps.shape != theta_pol_steps.shape or \
                theta_ret_steps.shape != pol_in.shape or \
                theta_ret_steps.shape != ret_in.shape or \
                theta_ret_steps.shape != timeobs.shape:
            raise ValueError('Calibration Sequence step arrays do not have the same shape')

        self.theta_pol_steps = theta_pol_steps * np.pi / 180
        self.theta_ret_steps = theta_ret_steps * np.pi / 180
        self.numsteps = self.theta_pol_steps.shape[0]
        self.py = py
        self.pol_in = pol_in
        self.ret_in = ret_in
        self.dark_in = dark_in
        self.delta_t = timeobs - np.min(timeobs)
        self.t_pol_0 = np.array([t_pol_0], dtype=np.float64)
        self.t_ret_0 = np.array([t_ret_0], dtype=np.float64)
        self.ret_0_h = np.array([ret_0_h], dtype=np.float64)
        self.dret_h = np.array([dret_h], dtype=np.float64)
        self.ret_0_45 = np.array([ret_0_45], dtype=np.float64)
        self.dret_45 = np.array([dret_45], dtype=np.float64)
        self.ret_0_r = np.array([ret_0_r], dtype=np.float64)
        self.dret_r = np.array([dret_r], dtype=np.float64)
        self.sequence_nums = [self.numsteps]
        self.Q_in = Q_in
        self.U_in = U_in
        self.V_in = V_in
        self.I_sys = np.array([I_sys], dtype=np.float64)

    @property
    def numdrawers(self):
        """The number of Drawers (AKA CS's) represented in this object

        Returns
        -------
        int
        """
        return len(self.sequence_nums)

    def __add__(self, other):
        """Concatenate two Calibration Sequences.

        Each Sequence retains its individual CU parameters

        Returns
        -------
        CalibrationSequence
            A new CalibrationSequence object
        """

        output = CalibrationSequence(np.append(self.theta_pol_steps, other.theta_pol_steps) * 180 / np.pi,
                                     np.append(self.theta_ret_steps, other.theta_ret_steps) * 180 / np.pi,
                                     np.append(self.pol_in, other.pol_in),
                                     np.append(self.ret_in, other.ret_in),
                                     np.append(self.dark_in, other.dark_in),
                                     np.append(self.delta_t, other.delta_t))

        output.t_pol_0 = np.append(self.t_pol_0, other.t_pol_0)
        output.t_ret_0 = np.append(self.t_ret_0, other.t_ret_0)
        output.I_sys = np.append(self.I_sys, other.I_sys)
        output.ret_0_h = np.append(self.ret_0_h, other.ret_0_h)
        output.dret_h = np.append(self.dret_h, other.dret_h)
        output.ret_0_45 = np.append(self.ret_0_45, other.ret_0_45)
        output.dret_45 = np.append(self.dret_45, other.dret_45)
        output.ret_0_r = np.append(self.ret_0_r, other.ret_0_r)
        output.dret_r = np.append(self.dret_r, other.dret_r)
        output.sequence_nums = self.sequence_nums + other.sequence_nums
        output.Q_in = self.Q_in
        output.U_in = self.U_in
        output.V_in = self.V_in

        return output

    def init_with_dresser(self, dresser):
        """Initialize CU parameters from a Dresser

        Really all this does is automatically set the CS configuration variables (theta_pol_steps, etc.) and then
        initialize zero arrays of the appropriate length for all of the parameters that will be fit.

        Parameters
        ----------
        dresser : `Data.Dresser`
            Object containing one or more Drawers of SoCCs.
        """
        for attr in ['theta_pol_steps', 'theta_ret_steps', 'pol_in', 'ret_in', 'dark_in', 'delta_t', 'numsteps']:
            setattr(self, attr, getattr(dresser, attr))

        self.py = 0.
        self.Q_in = 0.
        self.U_in = 0.
        self.V_in = 0.
        self.theta_pol_steps *= np.pi / 180
        self.theta_ret_steps *= np.pi / 180
        self.sequence_nums = dresser.drawer_step_list
        self.t_pol_0 = 1 + np.zeros(dresser.numdrawers, dtype=np.float64)
        self.t_ret_0 = 1 + np.zeros(dresser.numdrawers, dtype=np.float64)
        self.ret_0_h = np.zeros(dresser.numdrawers, dtype=np.float64)
        self.dret_h = np.zeros(dresser.numdrawers, dtype=np.float64)
        self.ret_0_45 = np.zeros(dresser.numdrawers, dtype=np.float64)
        self.dret_45 = np.zeros(dresser.numdrawers, dtype=np.float64)
        self.ret_0_r = np.zeros(dresser.numdrawers, dtype=np.float64)
        self.dret_r = np.zeros(dresser.numdrawers, dtype=np.float64)
        self.I_sys = 1. + np.zeros(dresser.numdrawers, dtype=np.float64)

    def load_pars_from_dict(self, params):
        """Update CU Model parameters based on a dictionary of the same

        This doesn't do any checks because it is assumed that `FittingFramework.verify_fit_mode` will have already
        done so.

        Probably don't call this directly

        Parameters
        ----------
        params : dict
            CU Model parameter key: value pairs
        """
        self.Q_in = params['Q_in']
        self.U_in = params['U_in']
        self.V_in = params['V_in']

        for i in range(self.numdrawers):
            self.t_pol_0[i] = params['t_pol_CS{:02n}'.format(i)]
            self.t_ret_0[i] = params['t_ret_CS{:02n}'.format(i)]

            self.I_sys[i] = params['I_sys_CS{:02n}'.format(i)]

            self.ret_0_h[i] = params['ret0h_CS{:02n}'.format(i)]
            self.dret_h[i] = params['dreth_CS{:02n}'.format(i)]
            self.ret_0_45[i] = params['ret045_CS{:02n}'.format(i)]
            self.dret_45[i] = params['dret45_CS{:02n}'.format(i)]
            self.ret_0_r[i] = params['ret0r_CS{:02n}'.format(i)]
            self.dret_r[i] = params['dretr_CS{:02n}'.format(i)]

    def set_py_from_database(self, wavelength):
        """Update the transmission coefficient in the y direction (py)

        Linear interpolation is used to match the input wavelength. py values outside of the database wavelength range
        are simply extended from the min/max database values.
        """
        wave, py = np.loadtxt(pkg_resources.resource_filename('dkist_processing_pac', 'data/py_table.txt'), unpack=True)
        self.py = np.interp(wavelength, wave, py)

        print('{}: py = {:7.6f} at {:4n} nm'.format(tag(), self.py, wavelength))

    @property
    def t_pol(self):
        """Transmission of the polarizer at each step in the CS

        If multiple CS's exist then the transmission will be valid for the entire suite.

        Returns
        -------
        numpy.ndarray
            Array of shape (N,)
        """
        t_pol = np.ones(self.numsteps)
        id1 = 0
        id2 = 0

        for i in range(self.numdrawers):
            id2 += self.sequence_nums[i]
            t_pol[id1:id2] = self.t_pol_0[i]
            id1 = id2

        return t_pol

    @property
    def t_ret(self):
        """Transmission of the retarder at each step in the CS

        If multiple CS's exist then the transmission will be valid for the entire suite.

        Returns
        -------
        numpy.ndarray
            Array of shape (N,)
        """
        t_ret = np.ones(self.numsteps)
        id1 = 0
        id2 = 0

        for i in range(self.numdrawers):
            id2 += self.sequence_nums[i]
            t_ret[id1:id2] = self.t_ret_0[i]
            id1 = id2

        return t_ret

    @property
    def S_in(self):
        """The Stokes vector incident on the Calibration Unit

        NOTE that this does not include the effects of M12. S_in is parametrized via:

        S_in = I_sys * [1, Q_in, U_in, V_in]

        Returns
        -------
        numpy.ndarray
            Array of shape (N,4)
        """
        S_in = np.zeros((self.numsteps, 4), dtype=np.float64)
        id1 = 0
        id2 = 0

        for i in range(self.numdrawers):
            id2 += self.sequence_nums[i]

            S_in[id1:id2] = self.I_sys[i] * np.array([[1., self.Q_in, self.U_in, self.V_in]] * self.sequence_nums[i])
            id1 = id2

        return S_in

    @property
    def ret_h(self):
        """The computed retardance at each step of the CS.

        If multiple CS's exist then ret will be valid for the entire suite.

        Returns
        -------
        numpy.ndarray
            Array of shape (N,)
        """
        ret = np.zeros(self.numsteps)
        id1 = 0
        id2 = 0

        for i in range(self.numdrawers):
            id2 += self.sequence_nums[i]
            ret[id1:id2] = self.ret_0_h[i] + self.dret_h[i] * self.delta_t[id1:id2]
            id1 = id2

        return ret

    @property
    def ret_45(self):
        """The computed retardance at each step of the CS.

        If multiple CS's exist then ret will be valid for the entire suite.

        Returns
        -------
        numpy.ndarray
            Array of shape (N,)
        """
        ret = np.zeros(self.numsteps)
        id1 = 0
        id2 = 0

        for i in range(self.numdrawers):
            id2 += self.sequence_nums[i]
            ret[id1:id2] = self.ret_0_45[i] + self.dret_45[i] * self.delta_t[id1:id2]
            id1 = id2

        return ret

    @property
    def ret_r(self):
        """The computed retardance at each step of the CS.

        If multiple CS's exist then ret will be valid for the entire suite.

        Returns
        -------
        numpy.ndarray
            Array of shape (N,)
        """
        ret = np.zeros(self.numsteps)
        id1 = 0
        id2 = 0

        for i in range(self.numdrawers):
            id2 += self.sequence_nums[i]
            ret[id1:id2] = self.ret_0_r[i] + self.dret_r[i] * self.delta_t[id1:id2]
            id1 = id2

        return ret

    @property
    def pol_mat(self):
        """The set of Mueller matrices for the polarizer.

        Returns
        -------
        numpy.ndarray
            Array of shape (N, 4, 4)
        """
        pol_mat = np.empty((self.numsteps, 4, 4), dtype=np.float64)
        t_pol = self.t_pol

        for i in range(self.numsteps):
            if self.pol_in[i]:
                rot_in = Data.rotation(self.theta_pol_steps[i])
                rot_out = Data.rotation(-self.theta_pol_steps[i])

                pol_mat[i] = rot_out @ Data.polarizer(t_pol[i], self.py) @ rot_in
            else:
                pol_mat[i] = np.diag(np.ones(4, dtype=np.float64))

        return pol_mat

    @property
    def ret_mat(self):
        """The set of Mueller matrices for the retarder

        Returns
        -------
        numpy.ndarray
            Array of shape (N, 4, 4)
        """
        ret_mat = np.empty((self.numsteps, 4, 4), dtype=np.float64)
        t_ret = self.t_ret
        ret_h = self.ret_h
        ret_45 = self.ret_45
        ret_r = self.ret_r

        for i in range(self.numsteps):
            if self.ret_in[i]:
                rot_in = Data.rotation(self.theta_ret_steps[i])
                rot_out = Data.rotation(-self.theta_ret_steps[i])
                ret_mat[i] = rot_out @ Data.elliptical_retarder(t_ret[i], ret_h[i], ret_45[i], ret_r[i]) @ rot_in
            else:
                ret_mat[i] = np.diag(np.ones(4, dtype=np.float64))

        return ret_mat

    @property
    def dark_mat(self):
        dark_mat = np.empty((self.numsteps, 4, 4), dtype=np.float32)
        for i in range(self.numsteps):
            if self.dark_in[i]:
                dark_mat[i] = np.zeros((4,4), dtype=np.float32)
            else:
                dark_mat[i] = np.diag(np.ones(4, dtype=np.float32))

        return dark_mat

    @property
    def CM(self):
        """Set of Mueller matrices for the entire CU.

        Returns
        -------
        numpy.ndarray
            Array of shape (N, 4, 4)
        """

        return self.ret_mat @ self.pol_mat @ self.dark_mat



import numpy as np
from dkist_processing_pac import Data, CUModel, TelescopeModel, generic, FittingFramework
from dkist_processing_pac.tag import tag

def prep_fit(dirlist, suffix='.FITS', telescope_db=None, fit_mode='baseline', init_set='default', fit_TM=True,
             skip_darks=True):
    """Do the setup of a fit and return the relevant variables (Dresser, CS, TM, and params)

    The purpose is to save some console time during debugging.
    """
    DRSR = Data.Dresser()

    for drwr_loc in dirlist:
        # fitaux is a special directory that can be generated automatically for saving fit statistics
        #  the [-6:] is a poor-man's regex; ensure that fitaux is at the end of the dirname. We don't want to stop
        #  someone from using fitaux_is_my_dir_for_some_reason as a location to save fit results
        if drwr_loc[-6:] == 'fitaux':
            continue
        tmp_drawer = Data.Drawer(drwr_loc, suffix, skip_darks=skip_darks)
        DRSR.add_drawer(tmp_drawer)

        print('{}: loaded polcal data from {}'.format(tag(), drwr_loc))

    mode_opts = generic.init_fit_mode(fit_mode, DRSR.wavelength, fit_TM, init_set=init_set)
    global_transmission = mode_opts['switches']['global_transmission']

    # We need to initialize a dummy CU model
    dummy = np.zeros(1)
    CS = CUModel.CalibrationSequence(*([dummy] * 6))
    CS.init_with_dresser(DRSR)  # And now actually fill it with the correct configurations

    TM = TelescopeModel.TelescopeModel(DRSR.azimuth, DRSR.elevation, DRSR.table_angle)

    print('{}: updating polarizer py value'.format(tag()))
    wave = DRSR.wavelength
    CS.set_py_from_database(wave)

    print('{}: loading telescope database from {}'.format(tag(), telescope_db))
    if telescope_db is None:
        telescope_db = generic.get_default_telescope_db()
    mean_time = 0.5 * (DRSR.date_bgn + DRSR.date_end)
    TM.load_from_database(telescope_db, mean_time, wave)

    numx, numy, numl = DRSR.shape
    if fit_TM:
        print('{}: shape of Dresser data: ({}, {}, {})'.format(tag(), numx, numy, numl))
    else:
        print('{}: shape of Drawer data: ({}, {}, {})'.format(tag(), numx, numy, numl))
    print('{}: number of modulator states: {}'.format(tag(), DRSR.nummod))
    print('{}: number of CS steps: {}'.format(tag(), DRSR.numsteps))

    nummod = DRSR.nummod
    I_clear = DRSR.I_clear
    params = FittingFramework.generate_parameters(I_clear, TM, CS, mode_opts, nummod, fit_TM=fit_TM)

    return DRSR, CS, TM, mode_opts, params



import os
import copy
import multiprocessing as mp
import pickle
import time
import numpy as np
import lmfit
from queue import Empty as QueueEmptyError
from dkist_processing_pac import TelescopeModel, CUModel, Data, generic
from dkist_processing_pac.tag import tag, is_mp
from dkist_processing_pac.data import CONSTANTS
from dkist_processing_pac.data.required_params import elliptical

try:
    TERMCOL = int(os.popen('stty size', 'r').read().split()[1])
except:
    TERMCOL = 80

def run_fits(dirlist, suffix='.FITS', skip_darks=True, **kwargs):
    DRSR = Data.Dresser()

    for drwr_loc in dirlist:
        # fitaux is a special directory that can be generated automatically for saving fit statistics
        #  the [-6:] is a poor-man's regex; ensure that fitaux is at the end of the dirname. We don't want to stop
        #  someone from using fitaux_is_my_dir_for_some_reason as a location to save fit results
        if drwr_loc[-6:] == 'fitaux':
            continue
        tmp_drawer = Data.Drawer(drwr_loc, suffix, skip_darks=skip_darks)
        DRSR.add_drawer(tmp_drawer)

        print('{}: loaded polcal data from {}'.format(tag(), drwr_loc))

    return run_core(DRSR, **kwargs)

def run_core(dresser, telescope_db='telescope_db.txt', fit_mode='baseline', init_set='default',
             fit_TM=True, method='leastsq', pickle_path=None, threads=1, super_name='',
             **fit_kwargs):
    """Top-level function to ingest data and setup fitting of Telescope and CU parameters.

    First read all PolCal data from files, then create corresponding CU and Telescope objects, and, finally, kick off
    the fit.

    Data are expected to live in a nested directory structure where each sub-directory contains data from a single
    Calibration Sequence. Any number of CS's are allowed, and a subset of the sub-directories can be selected via the
    numCS parameter.

    Information related to the time and CU and Telescope states are read directly from the FITS headers of the data.
    The Telescope Model mirror parameters are pulled from the provided database, but the CU model is initialized with
    no assumptions.

    Parameters
    ----------
    dirlist : list
       List of directories from which to load Drawer data. Each directory should contain a single Drawer.

    suffix : str
        Filter to select specific files. File mask is top_level_dir/sub_dir/*suffix

    telescope_db : str
        Path to the database of Telescope Model mirror parameters. See TelescopeModel.TelescopeModel.load_from_database
        for more information.

    fit_mode : str
        The PA&C fit recipe to use for fitting

    fit_TM : bool
        If True then vary the M36 mirror parameters during the fit

    skip_darks : bool
        If True (default) then don't include and dark steps of the CS in fitting

    pickle_path : str
        Location of directory within which to save fitting statistics for later analysis

    method : str
        Fitting method to use. Valid values are:
        - `'differential_evolution'`: differential evolution (default)
        - `'leastsq'`: Levenberg-Marquardt
        - `'least_squares'`: Least-Squares minimization, using Trust Region Reflective method by default
        - `'brute'`: brute force method
        - '`nelder`': Nelder-Mead
        - `'lbfgsb'`: L-BFGS-B
        - `'powell'`: Powell
        - `'cg'`: Conjugate-Gradient
        - `'newton'`: Newton-Congugate-Gradient
        - `'cobyla'`: Cobyla
        - `'tnc'`: Truncate Newton
        - `'trust-ncg'`: Trust Newton-Congugate-Gradient
        - `'dogleg'`: Dogleg
        - `'slsqp'`: Sequential Linear Squares Programming

        For more details on the fitting methods please refer to the
        `SciPy docs <http://docs.scipy.org/doc/scipy/reference/optimize.html>`__.

    fit_kwargs : dict
        Options to be passed to the minimizer being used

    Returns
    -------
    `Data.CUModelParams`
        Object containing the best fit CU parameters

    `Data.TMModelParams`
        Object containing the best fit TM parameters

    `TelescopModel.TelescopeModel`
        Best fit Telescope Model

    Notes
    -----
    The shape of the input data controls how much spatial information is extracted from the fits; a separate fit is
    performed for each pixel in the input data. In many cases the input shape will be (1, 1), in which case no spatial
    information is extracted. The other extreme is to do no pre-binning of the data and fit a different set of
    parameters (and modulation matrices) for every single pixel of the instrument detector, although this will probably
    take a very long time.
    """

    mode_opts = generic.init_fit_mode(fit_mode, dresser.wavelength, fit_TM, init_set=init_set)
    generic.verify_fit_mode(mode_opts, fit_TM)
    use_T = mode_opts['switches']['use_T']
    use_M12 = mode_opts['switches']['use_M12']
    global_transmission = mode_opts['switches']['global_transmission']
    global_retardance = mode_opts['switches']['global_retardance']

    # We need to initialize a dummy CU model
    dummy = np.zeros(1)
    CS = CUModel.CalibrationSequence(*([dummy] * 6))
    CS.init_with_dresser(dresser)  # And now actually fill it with the correct configurations

    TM = TelescopeModel.TelescopeModel(dresser.azimuth, dresser.elevation, dresser.table_angle)

    print('{}: updating polarizer py value'.format(tag()))
    wave = dresser.wavelength
    CS.set_py_from_database(wave)

    print('{}: loading telescope database from {}'.format(tag(), telescope_db))
    mean_time = 0.5 * (dresser.date_bgn + dresser.date_end)
    TM.load_from_database(telescope_db, mean_time, wave)

    numx, numy, numl = dresser.shape
    if fit_TM:
        print('{}: shape of Dresser data: ({}, {}, {})'.format(tag(), numx, numy, numl))
    else:
        print('{}: shape of Drawer data: ({}, {}, {})'.format(tag(), numx, numy, numl))
    print('{}: number of modulator states: {}'.format(tag(), dresser.nummod))
    print('{}: number of CS steps: {}'.format(tag(), dresser.numsteps))

    print('{}: fitting CU parameters'.format(tag()))
    CU_params = Data.CUModelParams(numx, numy, numl, dresser.numdrawers, dresser.nummod,
                                   fit_mode=fit_mode, init_set=init_set,
                                   header=dresser.drawers[0].data_list[0][0].header)
    TM_params = Data.TelescopeModelParams(numx, numy, numl, header=dresser.drawers[0].data_list[0][0].header)

    if threads > 1:
        print('{}: starting fits with {} threads'.format(tag(), threads))
        o = multiproc_drawer_fit(dresser, TM, CS, mode_opts, CU_params, TM_params, numx, numy, numl, int(threads), pickle_path,
                                 method=method, fit_TM=fit_TM, use_T=use_T, use_M12=use_M12, super_name=super_name,
                                 global_transmission=global_transmission, global_retardance=global_retardance,
                                 **fit_kwargs)
    else:
        o = singleproc_drawer_fit(dresser, TM, CS, mode_opts, CU_params, TM_params, numx, numy, numl, pickle_path,
                                  method=method, fit_TM=fit_TM, use_T=use_T, use_M12=use_M12,
                                  global_transmission=global_transmission, global_retardance=global_retardance,
                                  **fit_kwargs)

    return CU_params, TM_params, o[-1]

def singleproc_drawer_fit(Dresser, TM, CS, mode_opts, CU_params, TM_params, numx, numy, numl, pickle_path, fit_TM=False,
                          **kwargs):

    t1 = time.time()
    I_clear = Dresser.I_clear
    for i in range(numx):
        for j in range(numy):
            for k in range(numl):
                if fit_TM:
                    print('{}:  using StoCCinG at position = ({}, {}, {})'.format(tag(), i, j, k))
                else:
                    print('{}:  using SoCC at position = ({}, {}, {})'.format(tag(), i, j, k))
                I_cal, I_unc = Dresser[i, j, k]
                try:
                    TMf, CSf, minimizer, fit_out = fit_parameters(I_cal, I_unc, I_clear, copy.deepcopy(TM),
                                                                  copy.deepcopy(CS), mode_opts, fit_TM=fit_TM, **kwargs)
                except ValueError as e:
                    if 'NaN values detected' in str(e):
                        print('{}: NaN values detected either in input or fitting calculations. Assigning garbage results.'.format(tag()))
                        fit_out = None
                        TMf = None
                        minimizer = None
                    else:
                        raise

                CU_params[i, j, k] = fit_out
                TM_params[i, j, k] = TMf

                if pickle_path is not None:
                    pickle_name = '{}/fitobj_{:04n}{:04n}{:04n}.pkl'.format(pickle_path, i, j, k)
                    print('{}: saving auxiliary fitting data to {}'.format(tag(), pickle_name))
                    with open(pickle_name, 'wb') as f:
                        pickle.dump(minimizer, f)
                        pickle.dump(fit_out, f)

    print('{}: all fits took {:.1f} s'.format(tag(), time.time() - t1))

    return CU_params, TM_params, TMf

def multiproc_drawer_fit(Dresser, TM, CS, mode_opts, CU_params, TM_params, numx, numy, numl, numthread, pickle_path,
                         super_name='', **kwargs):

    # How many fits will be run on each thread?
    num_points = numx * numy * numl
    fits_per_thread = int(num_points / numthread)
    if num_points / numthread % 1 > 0:
        fits_per_thread += 1

    # These loops set up a list of mp processes. They have not been started.
    jobs = []
    lock = mp.Lock()
    queue = mp.Queue()
    thread_pos_inputs = [[] for _ in range(numthread)]
    f = 0
    print('{}: constructing jobs'.format(tag()))
    while f < num_points:
        for k in range(numthread):
            pos = np.unravel_index(f, (numx, numy, numl))
            thread_pos_inputs[k].append(pos)
            f += 1
            if f >= num_points:
                break

    print('{}: {} threads will be started with <= {} fits per thread'.format(tag(), numthread, fits_per_thread))
    if super_name:
        name_pre = '{}: '.format(super_name)
    else:
        name_pre = ''
    for k in range(numthread):
        print('{}: thread {} will fit positions {}'.format(tag(), k, thread_pos_inputs[k]))
        args = [Dresser, copy.deepcopy(TM), copy.deepcopy(CS), mode_opts, thread_pos_inputs[k],
                pickle_path, lock, queue]
        jobs.append(mp.Process(target=do_SoCC_fit,
                               args=args,
                               kwargs=kwargs,
                               name='{}Thread {} X'.format(name_pre, k)))

    # Now start the jobs
    t1 = time.time()
    for p in jobs:
        p.start()

    # And now we loop until all outputs have been captured
    num_completed = 0
    num_timeouts = 0
    status_str = '{:}: FIT STATUS: {:} of {:} points fit in {:.1f} s. There are {:} threads still active and the Queue contains {:} result(s)'
    try:
        while num_completed < num_points:
            print(status_str.format(tag(), num_completed, num_points, time.time() - t1,
                                    [x.is_alive() for x in jobs].count(True), queue.qsize()))
            try:
                # queue.get() blocks until there is something in the queue
                ii, jj, kk, fo, TMf = queue.get(timeout=CONSTANTS['multiproc_queue_timeout'])
            except QueueEmptyError:
                # For reasons that I can't figure out sometimes the queue hangs. Thus the timeout. Here we catch the
                # resulting error and simply print a message. The loop will repeat and usually just calling queue.get() again
                # will kick the queue out of whatever funk it was in.
                print('{:}: FIT STATUS: Queue timeout ({:}s) reached'.format(tag(), CONSTANTS['multiproc_queue_timeout']))
                num_timeouts += 1
                if num_timeouts >= CONSTANTS['multiproc_max_timeouts']:
                    print('{}: FIT STATUS: Max timeouts ({}) reached. Breaking out of fit loop, '
                          'this may cause some errors.'.format(tag(), CONSTANTS['multiproc_max_timeouts']))
                    for p in jobs:
                        p.terminate()
                    break
            else:
                CU_params[ii, jj, kk] = fo
                TM_params[ii, jj, kk] = TMf
                num_completed += 1

            time.sleep(0.05)  # This is possibly unnecessary due to the block of queue.get(), but it's left here
                              #  because a) setting the *_params requires file IO, and b) just to be paranoid.
                              #  That said, the synchronus nature of this loop (caused by .get() blocking) means
                              #  this wait can probably safely be removed.
    except KeyboardInterrupt:
        print("{}: FIT STATUS: KEYBOARD INTERRUPT! There are {} threads still active and the Queue contains {} "
              "result(s). You're going to see a lot of errors".format(tag(),
                                                                      [x.is_alive() for x in jobs].count(True),
                                                                      queue.qsize()))
        print('{}: FIT STATUS: Only {} of {} fits completed. '
              'Now attempting to join the remaining threads.'.format(tag(), num_completed, num_points))

    print('{}: done with loop in {:.1f} s'.format(tag(), time.time() - t1))
    print('{}: There are still {} active threads and the queue has {} objects in it'.format(tag(), [x.is_alive() for x in jobs].count(True),
                                                                                            queue.qsize()))
    # Next, make sure we wait for all jobs to complete before moving on. Because of how we constructed the previous
    #  while loop, this might not matter at all.
    try:
        for p in jobs:
            p.join()
    except KeyboardInterrupt:
        print('{}: FIT STATUS: KEYBOARD INTERRUPT! Join() was stopped. Goodbye and good luck.'.format(tag()))

    # Finally, we'll cleanup the Queue. This is not only good practice, but also an attempt to address some reports
    #  from the field that resources are not released and can eventually clog a system if the user is calling run_fits()
    #  many/multiple times in the same python session.
    print('{}: Cleaning up mp Queue'.format(tag()))
    try:
        queue.close()
        queue.join_thread()
    except KeyboardInterrupt:
        print('{}: FIT STATUS: KEYBOARD INTERRUPT! The Queue cleanup was stopped. You monster! '
              'Fitting results are unaffected, but resources might not have been freed.'.format(tag()))

    print('{}: all fits took {:.1f} s'.format(tag(), time.time() - t1))

    return CU_params, TM_params, TM

def do_SoCC_fit(Dresser, TM, CS, mode_opts, pos_list, pickle_path, lock, queue, **kwargs):


    I_clear = Dresser.I_clear
    proc = mp.current_process()
    for pos in pos_list:
        i, j, k = pos
        with lock:
            I_cal, I_unc = Dresser[pos]
        proc.name = proc.name.replace('X', 'POS ({}, {}, {})'.format(i, j, k))
        print('{}: Starting fit of SoCC at position = ({}, {}, {})'.format(tag(), i, j, k))
        try:
            TMf, CSf, minimizer, fit_out = fit_parameters(I_cal, I_unc, I_clear, TM, CS, mode_opts, **kwargs)
        except ValueError as e:
            if 'NaN values detected' in str(e) or 'DLASCL' in str(e):
                print(
                    '{}: NaN values detected either in input or fitting calculations. Assigning garbage results.'.format(
                        tag()))
                fit_out = None
                TMf = None
                minimizer = None
            else:
                raise

        # A potential cause of the queue.get() hangs discussed above in multiproc_drawer_fit is that the queue.put()
        # call will hang indefinitely if the queue is "full" in some sense that is different than what is discussed
        # in the multiprocessing documentation. Unfortunately there seems to be no way to check for this condition. Thus
        # here we make sure that we don't put() to the queue until there is a sufficiently small number of items in the
        # queue. The idea is that we can set the multiproc_queue_objectlimit to be very small so that we'll never
        # overload the queue. The way multiproc_drawer_fit is constructed, the queue should be depopulated almost as
        # fast as items are pushed to it. Thus, this check is primarily designed to avoid a large number of threads
        # pushing their results all at once.
        while queue.qsize() >= CONSTANTS['multiproc_queue_objectlimit']:
            print('{}: FIT STATUS: Queue object limit ({}) reached. '
                  'This process is waiting to put its results on the queue.'.format(tag(),
                                                                                    CONSTANTS['multiproc_queue_objectlimit']))
            time.sleep(5)

        queue.put([i, j, k, fit_out, TMf])

        if pickle_path is not None:
            with lock:
                pickle_name = '{}/fitobj_{:04n}{:04n}{:04n}.pkl'.format(pickle_path, i, j, k)
                print('{}: saving auxiliary fitting data to {}'.format(tag(), pickle_name))
                with open(pickle_name, 'wb') as f:
                    pickle.dump(minimizer, f)
                    pickle.dump(fit_out, f)

        proc.name = proc.name.replace('POS ({}, {}, {})'.format(i, j, k), 'X')
        del I_cal

    queue.close()
    return

def generate_parameters(I_clear: np.ndarray,
                        TM: TelescopeModel.TelescopeModel,
                        CS: CUModel.CalibrationSequence,
                        mode_opts: dict,
                        nummod: int,
                        fit_TM: bool = True) -> lmfit.Parameters:
    """Create a Parameters set for an arbitrarily complex set of CS data.

    The TelescopeModel parameters are hardwired, but the CS can contain any number of single CalibrationSequences and
    therefore any number of CS parameter sets.

    Parameters
    ----------
    I_clear : numpy.ndarray
        1D array containing the average clear flux for each CS used

    TM : TelescopeModel.TelescopeModel
        An object describing the telescope configuration at each step in the CS

    CS : CUModel.CalibrationSequence
        An object describing the CU configuration at each step in the CS

    mode_opts : `configparser.ConfigParser`
        The options specified by a PA&C fitting recipe

    nummod : int
        The number of modulation states that will be fit. Used to define the modulation matrix variables.

    fit_TM : bool
        If True the Telescope Model parameters are allowed to vary in the fit

    global_tranmission : bool
        If True then all t_pol's will be the same and all t_ret's will be the same.

    global_retardance : bool
        If True then all 3 elliptical retardance parameters wll be the same for all CS's.

    Returns
    -------
    lmfit.Parameters
        Object containing information about the parameters to be used in the fit
    """

    params = lmfit.Parameters()

    ## Add telescope mirror parameters
    #
    telescope_pars = ['x34', 't34', 'x56', 't56']
    for tp in telescope_pars:
        if fit_TM:
            val = mode_opts[tp]['value']
        else:
            val = getattr(TM, tp)
        params.add(tp, value=val, vary=fit_TM,
                   min=mode_opts[tp]['min'],
                   max=mode_opts[tp]['max'])

    ## This will be a switch that defines how we treat the modulation matrix variables.
    ##  If ANY CU parameter has no bounds then we'll release the bounds on the modulation matrix as well.
    #
    CU_bounds = mode_opts['switches']['fit_bounds']

    ## Add CU Model parameters
    #
    # Add all CU parameters (except I_sys)
    for s in mode_opts.keys():
        if s == 'switches' or s == 'I_sys' or s in telescope_pars:
            continue

        if s in elliptical['global']:
            # Global parameters get a single value for all CSs
            params.add(s, **mode_opts[s])
        else:
            # These parameters are fit for every CS and thus we need a separate instance for each CS
            for i in range(CS.numdrawers):
                params.add('{}_CS{:02n}'.format(s, i), **mode_opts[s])

    # Now add I_sys. This is separate because I_sys is set relative to the actual data values
    if 'I_sys' in elliptical['global']:
        val = np.mean(I_clear)
        params.add('I_sys',
                   value=val,
                   min=val * mode_opts['I_sys']['min'],
                   max=val * mode_opts['I_sys']['max'],
                   vary=mode_opts['I_sys']['vary'])
    else:
        for i in range(CS.numdrawers):
            val = I_clear[i]
            params.add('I_sys_CS{:02n}'.format(i),
                       value=val,
                       min=val * mode_opts['I_sys']['min'],
                       max=val * mode_opts['I_sys']['max'],
                       vary=mode_opts['I_sys']['vary'])

    ## Set global transmission
    #
    if mode_opts['switches']['global_transmission']:
        # Kind of a hack; if global transmission is desired then we tie all transmissions to the first transmission
        pol_expr_str = '1. * t_pol_CS00'
        ret_expr_str = '1. * t_ret_CS00'
        for i in range(1, CS.numdrawers): # Start at 1 so we don't infinitely recurse on the "global" value
            params['t_pol_CS{:02n}'.format(i)].set(expr=pol_expr_str)
            params['t_ret_CS{:02n}'.format(i)].set(expr=ret_expr_str)

    ## Set global retardance values
    #   This is kind of a hack right now. If this ever makes it out of TelCal testing then I should modify how
    #    CalibrationSequence objects load parameters and change global-ness via the required_params.py file
    if mode_opts['switches']['global_retardance']:
        expr_str_h = '1. * ret0h_CS00'
        expr_str_45 = '1. * ret045_CS00'
        expr_str_r = '1. * ret0r_CS00'
        for i in range(1, CS.numdrawers):
            params['ret0h_CS{:02n}'.format(i)].set(expr=expr_str_h)
            params['ret045_CS{:02n}'.format(i)].set(expr=expr_str_45)
            params['ret0r_CS{:02n}'.format(i)].set(expr=expr_str_r)

    ## Add the modulation matrix variables
    #
    for m in range(nummod):
        for s in ['I', 'Q', 'U', 'V']:
            if CU_bounds:
                if s == 'I':
                    minval, maxval = 0.5, 1.5
                else:
                    minval, maxval = -1.5, 1.5
            else:
                minval, maxval = -np.inf, np.inf
            params.add('modmat_{}{}'.format(m, s), value=0, min=minval, max=maxval)

    # Now set modmat_0I to 1 and fix it. This ensures all the normalization stuff stays in I_sys
    params['modmat_0I'].set(value=1, vary=False)

    return params


def fit_parameters(I_cal, I_unc, I_clear, TM, CS, mode_opts, noprint=False,
                   method='differential_evolution', fit_TM=True, use_T=True, use_M12=True, global_transmission=False,
                   global_retardance=False, **fit_kwargs):
    """High-level fit-management function for determining the Telescope and CU parameters

    All this does is generate an appropriate parameter set and send it off to the minimizer.

    Due to the typically large number of parameters and complex chisq space, differential evolution is the recommended
    fitting method.

    Parameters
    ----------
    I_cal : numpy.ndarray
        Array of shape (M, N) containing the observed intensity values

    I_unc : numpy.ndarray
        Array of shape (M, N) containing uncertainties associated with I_cal

    I_clear : numpy.ndarray
        1D array containing the average clear flux for each separate CS used

    TM : TelescopeModel.TelescopeModel
        An object describing the telescope configuration at each step in the CS

    CS : CUModel.CalibrationSequence
        An object describing the CU configuration at each step in the CS

    mode_opts : `configparser.ConfigParser`
        The options specified by a PA&C fitting recipe

    method : str
        The fitting method to use. Valid values are:
        - `'differential_evolution'`: differential evolution (default)
        - `'leastsq'`: Levenberg-Marquardt
        - `'least_squares'`: Least-Squares minimization, using Trust Region Reflective method by default
        - `'brute'`: brute force method
        - '`nelder`': Nelder-Mead
        - `'lbfgsb'`: L-BFGS-B
        - `'powell'`: Powell
        - `'cg'`: Conjugate-Gradient
        - `'newton'`: Newton-Congugate-Gradient
        - `'cobyla'`: Cobyla
        - `'tnc'`: Truncate Newton
        - `'trust-ncg'`: Trust Newton-Congugate-Gradient
        - `'dogleg'`: Dogleg
        - `'slsqp'`: Sequential Linear Squares Programming

        For more details on the fitting methods please refer to the
        `SciPy docs <http://docs.scipy.org/doc/scipy/reference/optimize.html>`__.

    fit_TM : bool
        If True the Telescope Model parameters are allowed to vary in the fit

    fit_kwargs : dict
        Options to be passed to the minimizer being used

    noprint : bool
        If True don't print a live update of the free parameters, even if the fit is single-threaded

    Returns
    -------
    TelescopeModel.TelescopeModel
        Object containing the best fit Telescope Model

    CUModel.CalibrationSequence
        Object containing the best fit CU Model

    lmfit.Minimizer
        Minimizer object

    lmfit.MinimizerResult
        A container that holds the results of and stats about the fit
    """
    nummod = I_cal.shape[0]
    params = generate_parameters(I_clear, TM, CS, mode_opts, nummod, fit_TM=fit_TM)

    print('{}: initializing modulation matrix'.format(tag()))
    initialize_modulation_matrix(params, I_cal, TM, CS, use_T=use_T, use_M12=use_M12)

    print('{}: starting minimizer'.format(tag()))

    if not is_mp():
        print_header(params, noprint=noprint)

    modmat = np.zeros((I_cal.shape[0], 4), dtype=np.float64)
    t1 = time.time()
    mini = lmfit.Minimizer(compare_I, params, fcn_args=(I_cal, I_unc, TM, CS, modmat, t1),
                           fcn_kws={'use_T': use_T, 'use_M12': use_M12, 'noprint': noprint})
    fit_out = mini.minimize(method=method, params=params, **fit_kwargs)

    if not (is_mp() or noprint):
        print()

    print('{:}: minimization completed in {:4.1f} s. Chisq = {:8.2e}, redchi = {:8.2e}'.format(tag(), time.time() - t1,
                                                                                               fit_out.chisqr,
                                                                                               fit_out.redchi))

    return TM, CS, mini, fit_out


def compare_I(params, I_cal, I_unc, TM, CS, modmat, t1, use_T=True, use_M12=True, noprint=False):
    """The fitness/fitting function for fitting Telescope and CU parameters.

    First, a modulation matrix is determined and then used to generate some model output data. These data are then
    compared to the observed output data.

    Parameters
    ----------
    params : lmfit.Parameters
        Object containing information about the parameters to be used in the fit

    I_cal : numpy.ndarray
        Array of shape (M, N) containing the observed intensity for each modulator state (M) in each CS step (N).

    TM : TelescopeModel.TelescopeModel
       An object describing the telescope configuration at each step in the CS

    CS : CUModel.CalibrationSequence
        An object describing the CU configuration at each step in the CS

    t1 : float
        Time the fit was started, in seconds since the Unix Epoch.

    use_T : bool
        If True then include the M36 Mueller matrix

    use_M12 : bool
        If True then include the M12 Mueller matrix

    noprint : bool
        If True don't print a live update of the free parameters, even if the fit is single-threaded

    Returns
    -------
    numpy.ndarray
        Array of shape (M * N,) containing the difference between the observed and modeled data, normalized by the
        observed data. lmfit.minimize handles the squaring and summing internally.
    """
    parvals = params.valuesdict()
    TM.x34 = parvals['x34']
    TM.t34 = parvals['t34']
    TM.x56 = parvals['x56']
    TM.t56 = parvals['t56']

    CS.load_pars_from_dict(parvals)

    S = generate_S(TM, CS, use_T=use_T, use_M12=use_M12)
    #O = fit_modulation_matrix(I_cal, S)
    O = fill_modulation_matrix(parvals, modmat)

    I_mod = generate_model_I(O, S)

    diff = np.ravel((I_mod - I_cal) / I_unc)

    if not (is_mp() or noprint):
        print_status(params, diff, t1)

    return diff


def generate_S(TM, CS, use_T=True, use_M12=True):
    """Compute the Stokes vector immediately after the Calibration Unit

    Parameters
    ----------
    TM : TelescopeModel.TelescopeModel
        The TelescopeModel object used for fitting

    CS : CUModel.CalibrationSequence
        The CalibrationSequence object used for fitting

    use_T : bool
        If True then include the M36 Mueller matrix

    use_M12 : bool
        If True then include the M12 Mueller matrix

    Returns
    -------
    numpy.ndarray
        (4, N) array containing the CU output Stokes vector at each step in the CS
    """
    S = np.diag(np.ones(4, dtype=np.float64))
    if use_T:
        S = S @ TM.TM

    S = S @ CS.CM

    if use_M12:
        S = S @ TM.M12

    S = np.sum(S * CS.S_in[:, None, :], axis=2).T

    return S

def fit_modulation_matrix(I_cal, S):
    """Find a modulation matrix, O, that satisfies I = O T s.

    In our case the problem will most likely be massively over-constrained so we use a least-squares fit to find O.

    Parameters
    ----------
    M -> number of modulator states
    N -> number of steps in Calibration Sequence

    I_cal : numpy.ndarray
        Array of shape (M, N) containing the observed intensity values

    S : numpy.ndarray
        Array of shape (4, N) containing the model Stokes vector input to the instrument (i.e., after M6)

    Returns
    -------
    numpy.ndarray
        Array of shape (M, 4) containing the instrument modulation matrix.

    """
    # linalg.lstsq solves x for Ax = B, but in our case we want to solve O for I = O S.
    #  Fortunately we can use the identity Ax = B <-> x.T A.T = B.T and therefore recast our problem into the proper
    #  form: I.T = S.T O.T

    res = np.linalg.lstsq(S.T, I_cal.T, rcond=None)
    O = res[0].T

    #O /= np.max(O[:,0])
    O /= O[0,0]

    return O

def initialize_modulation_matrix(params, I_cal, TM, CS, use_T=True, use_M12=True):
    parvals = params.valuesdict()
    TM.x34 = parvals['x34']
    TM.t34 = parvals['t34']
    TM.x56 = parvals['x56']
    TM.t56 = parvals['t56']

    CS.load_pars_from_dict(parvals)

    S = generate_S(TM, CS, use_T=use_T, use_M12=use_M12)
    O = fit_modulation_matrix(I_cal, S)

    for m in range(O.shape[0]):
        for i, s in enumerate(['I', 'Q', 'U', 'V']):
            params['modmat_{}{}'.format(m,s)].set(value=O[m, i])

    # Just in case
    params['modmat_0I'].set(value=1)

def fill_modulation_matrix(parvals, modmat):

    nummod = modmat.shape[0]
    for m in range(nummod):
        for i, s in enumerate(['I', 'Q', 'U', 'V']):
            modmat[m, i] = parvals['modmat_{}{}'.format(m, s)]

    return modmat

def generate_model_I(O, S):
    """Given models for the telescope, CU, and modulation matrix, construct ideal output data.

    The "input" light is taken from the CalibrationSequence object and is probably unpolarized light ([1,0,0,0]).

    Parameters
    ----------
    O : numpy.ndarray
        Array of shape (M, 4) containing the instrument modulation matrix.

    S : numpy.ndarray
        Array of shape (4, N) containing the model Stokes vector input to the instrument (i.e., after M6)

    Returns
    -------
    numpy.ndarray
        Array of shape (M, N) containing the modeled intensity in each modulation state at each step of the CS.
    """

    I_mod = O @ S

    return I_mod


def print_status(params, diff, t1):
    """Print the current state of the fit.

    If the full suite of free parameters is too large to fit in a single terminal window then an appropriate number of
    free parameters will be hidden and their presence indicated with a '...'. If the current terminal has less than 25
    columns then this function will not stop wrapping. Use a bigger terminal!

    This is a separate function so that the logic to fit the message to the current terminal is isolated. To avoid a
    ton of system calls during fitting the terminal width is only computed on import.

    Parameters
    ----------
    params : lmfit.Parameters
        Object containing information about the parameters to be used in the fit

    diff : float
        Sqrt(chi^2). The minimization parameter used by the fit.

    t1 : float
        The time at which the fit was started (UNIX format)
    """
    status_str = '\r'

    next_width = 10
    truncated = False
    for v in params.keys():
        if params[v].vary:
            if next_width > TERMCOL - 25:
                truncated = True
                continue

            if np.log10(np.abs(params[v].value) + 1) >= 5:
                status_str += '{:>10.3e}'.format(params[v].value)
            else:
                status_str += '{:>10.3f}'.format(params[v].value)
            next_width += 10

    if truncated:
        status_str += '{:>5}'.format('...')

    status_str += '{:10.3e}'.format(np.sum(diff**2))

    dt = time.time() - t1
    time_unit = ' s'
    if dt > 120:
        dt /= 60
        time_unit = ' m'
    status_str += '{:8.1f}{:}'.format(dt, time_unit)

    print(status_str, end="", flush=True)

#TODO: Print initial values as well
def print_header(params, noprint=False):
    """Print fixed parameters and a header that defines columns in the free-parameter status message.

    If the full suite of free parameters is too large to fit in a single terminal window then an appropriate number of
    parameters will be hidden and their presence indicated with a '...'. If the current terminal has less than 25
    columns then this function will not stop wrapping. Use a bigger terminal!

    This is a separate function so that the logic to fit the message to the current terminal is isolated. The terminal
    width is computed once on import.

    Parameters
    ----------
    params : lmfit.Parameters
        Object containing information about the parameters to be used in the fit

    noprint : bool
        If True then the column headers for live updates of the free parameters will not be printed
    """
    fixed_str = 'Fixed parameters:\n'
    init_str = 'Starting values:\n'
    I_sys_str = 'I_sys starting values:\n'
    free_str = ''

    next_free_width = 10
    truncated = False
    for v in params.keys():
        if 'I_sys' in v:
            I_sys_str += '{:<10} : {:5.3f}\n'.format(v.replace('CS', ''), params[v].value)

        if params[v].vary:
            if 'I_sys' not in v:
                init_str += '{:<10} : {:5.3f}\n'.format(v.replace('CS', ''), params[v].value)
            if next_free_width > TERMCOL - 25:
                truncated = True
                continue

            free_str += '{:>10}'.format(v.replace('CS', ''))
            next_free_width += 10
        else:
            fixed_str += '{:<10} : {:5.3f}\n'.format(v.replace('CS', ''), params[v].value)

    if not noprint:
        I_sys_str += '\nFree parameters:'

        if truncated:
            free_str += '{:>5}'.format('...')

        free_str += '{:>10}{:>10}'.format('chisq', 't_elapsed')

    else:
        I_sys_str += '\nFree parameter live update suppressed'

    print(fixed_str)
    print(init_str)
    print(I_sys_str)
    if not noprint:
        print(free_str)



from unittest import TestCase
import numpy as np
from dkist_processing_pac.CUModel import CalibrationSequence
from dkist_processing_pac import Data

class TestInit(TestCase):

    def test_numsteps(self):

        psteps = rsteps = np.arange(4)
        pin = rin = din = np.array([1,1,0,0], dtype=np.bool)
        times = np.arange(4) * 1e4 + 1e4

        CS = CalibrationSequence(psteps, rsteps, pin, rin, din, times)
        self.assertEqual(CS.numsteps, 4)
        self.assertEqual(CS.sequence_nums, [4])

    def test_delta_t(self):

        psteps = rsteps = np.arange(4)
        pin = rin = din = np.array([1,1,0,0], dtype=np.bool)
        times = np.arange(4) * 1e3 + 4e4

        CS = CalibrationSequence(psteps, rsteps, pin, rin, din, times)
        np.testing.assert_array_equal(CS.delta_t, np.arange(4) * 1e3)

    def test_angle_conversion(self):

        psteps = rsteps = np.array([0, 60, 45, 90])
        pin = rin = din = np.array([1,1,0,0], dtype=np.bool)
        times = np.arange(4) * 1e3 + 4e4

        CS = CalibrationSequence(psteps, rsteps, pin, rin, din, times)
        np.testing.assert_array_equal(CS.theta_ret_steps, np.array([0, np.pi/3, np.pi/4, np.pi/2.]))
        np.testing.assert_array_equal(CS.theta_pol_steps, np.array([0, np.pi/3, np.pi/4, np.pi/2.]))

    def test_unequal_lengths(self):

        psteps = np.arange(4)
        rsteps = np.arange(5)
        pin = rin = din = np.array([1,1,1,0,0,0], dtype=np.bool)
        times = np.arange(4) * 1e3 + 1e4

        with self.assertRaises(ValueError):
            CS = CalibrationSequence(psteps, rsteps, pin, rin, din, times)

    def test_kwargs_conversion(self):

        psteps = rsteps = np.arange(4)
        pin = rin = din = np.array([1,1,0,0], dtype=np.bool)
        times = np.arange(4) * 1e4 + 1e4

        CS = CalibrationSequence(psteps, rsteps, pin, rin, din, times, ret_0_h=1, dret_h=3, ret_0_45=10., dret_45=30., ret_0_r=0.1, dret_r=0.3)
        np.testing.assert_array_equal(CS.ret_0_h, np.array([1.], dtype=np.float64))
        np.testing.assert_array_equal(CS.dret_h, np.array([3.], dtype=np.float64))
        np.testing.assert_array_equal(CS.ret_0_45, np.array([10.], dtype=np.float64))
        np.testing.assert_array_equal(CS.dret_45, np.array([30.], dtype=np.float64))
        np.testing.assert_array_equal(CS.ret_0_r, np.array([0.1], dtype=np.float64))
        np.testing.assert_array_equal(CS.dret_r, np.array([0.3], dtype=np.float64))

class TestAddition(TestCase):

    def test_concatenation(self):

        CS1 = CalibrationSequence(np.array([0,45]), np.array([90,0]), np.array([True, True]), np.array([False, True]),
                                  np.array([False, True]), np.array([50000, 60000]))
        CS2 = CalibrationSequence(np.array([90,180]), np.array([-90,180]), np.array([False, True]), np.array([True, True]),
                                  np.array([True, False]), np.array([55000, 65000]))
        CS = CS1 + CS2
        np.testing.assert_array_equal(CS.theta_pol_steps, np.array([0,45,90,180]) * np.pi / 180)
        np.testing.assert_array_equal(CS.theta_ret_steps, np.array([90,0,-90,180]) * np.pi / 180)
        np.testing.assert_array_equal(CS.pol_in, np.array([True, True, False, True]))
        np.testing.assert_array_equal(CS.ret_in, np.array([False, True, True, True]))
        np.testing.assert_array_equal(CS.delta_t, np.array([0, 10000, 0, 10000]))
        np.testing.assert_array_equal(CS.dark_in, np.array([False, True, True, False]))

    def test_kwargs(self):

        CS1 = CalibrationSequence(np.array([0,45]), np.array([90,0]), np.array([True, True]), np.array([False, True]),
                                  np.array([False, True]), np.array([50000, 60000]), ret_0_h=1, dret_h=10)
        CS2 = CalibrationSequence(np.array([90,180]), np.array([-90,180]), np.array([False, True]), np.array([True, True]),
                                  np.array([True, False]), np.array([55000, 65000]), ret_0_h=2, dret_h=11)
        CS = CS1 + CS2
        np.testing.assert_array_equal(CS.ret_0_h, np.array([1,2]))
        np.testing.assert_array_equal(CS.dret_h, np.array([10,11]))
        self.assertEqual(CS.sequence_nums, [2,2])

    def test_new_object(self):

        CS1 = CalibrationSequence(np.array([0,45]), np.array([90,0]), np.array([True, True]), np.array([False, True]),
                                  np.array([False, True]), np.array([50000, 60000]))
        CS2 = CalibrationSequence(np.array([90,180]), np.array([-90,180]), np.array([False, True]), np.array([True, True]),
                                  np.array([True, False]), np.array([55000, 65000]))
        CS = CS1 + CS2
        self.assertIsNot(CS, CS1)
        self.assertIsNot(CS, CS2)

class TestComponents(TestCase):

    def test_single_ret(self):

        psteps = rsteps = np.arange(5)
        pin = rin = din = np.array([1,1,0,0,1], dtype=np.bool)
        times = np.arange(5) * 1e3 + 4e4

        CS = CalibrationSequence(psteps, rsteps, pin, rin, din, times, ret_0_h=4, dret_h=1, ret_0_45=1, dret_45=2, ret_0_r=1, dret_r=-1)
        np.testing.assert_array_equal(CS.ret_h, np.arange(5) * 1e3 * 1 + 4)
        np.testing.assert_array_equal(CS.ret_45, np.arange(5) * 1e3 * 2 + 1)
        np.testing.assert_array_equal(CS.ret_r, np.arange(5) * 1e3 * -1 + 1)

    def test_multi_ret(self):

        CS1 = CalibrationSequence(np.array([0,45]), np.array([90,0]), np.array([True, True]), np.array([False, True]),
                                  np.array([False, True]), np.array([50000, 60000]), ret_0_h=1, dret_h=10, ret_0_45=10, dret_45=1, ret_0_r=1, dret_r=1)
        CS2 = CalibrationSequence(np.array([90,180]), np.array([-90,180]), np.array([False, True]), np.array([True, True]),
                                  np.array([True, False]), np.array([55000, 65000]), ret_0_h=2, dret_h=11, ret_0_45=20, dret_45=1, ret_0_r=2, dret_r=22)
        CS = CS1 + CS2
        np.testing.assert_array_equal(CS.ret_h, np.array([1, 100001, 2, 110002]))
        np.testing.assert_array_equal(CS.ret_45, np.array([10, 10010, 20, 10020]))
        np.testing.assert_array_equal(CS.ret_r, np.array([1, 10001, 2, 220002]))

    def test_pol_mat(self):

        CS = CalibrationSequence(np.array([0,45,90]), np.array([90,0,0]), np.array([True, False,True]),
                                 np.array([False, False, False]), np.array([False, False, False]),
                                 np.array([50000, 60000, 70000]))

        truth1 = Data.polarizer(1, 0)
        truth2 = np.diag(np.ones(4))
        truth3 = Data.rotation(-np.pi / 2) @ Data.polarizer(1, 0) @ Data.rotation(np.pi / 2)

        self.assertEqual(CS.pol_mat.shape, (3,4,4))
        np.testing.assert_array_equal(CS.pol_mat[0], truth1)
        np.testing.assert_array_equal(CS.pol_mat[1], truth2)
        np.testing.assert_array_equal(CS.pol_mat[2], truth3)

    def test_ret_mat(self):

        CS = CalibrationSequence(np.array([0,45,90]), np.array([90,0,45]), np.array([False, False, False]),
                                 np.array([False, True, True]), np.array([False, False, False]),
                                 np.array([1, 2, 3]), ret_0_h=0, dret_h=1, ret_0_45=1, dret_45=2, ret_0_r=1, dret_r=2)

        truth1 = np.diag(np.ones(4))
        truth2 = Data.elliptical_retarder(1, 1., 3., 3.)
        truth3 = Data.rotation(-np.pi/4) @ Data.elliptical_retarder(1, 2., 5., 5.) @ Data.rotation(np.pi / 4)

        self.assertEqual(CS.ret_mat.shape, (3,4,4))
        np.testing.assert_array_equal(CS.ret_mat[0], truth1)
        np.testing.assert_array_equal(CS.ret_mat[1], truth2)
        np.testing.assert_array_equal(CS.ret_mat[2], truth3)

    def test_CM(self):

        CS1 = CalibrationSequence(np.array([0, 45]), np.array([90, 45]),
                                  np.array([True, False]), np.array([True, False]), np.array([False, False]),
                                  np.array([1, 2]), ret_0_h=1, dret_h=1, ret_0_45=1, dret_45=2, ret_0_r=0, dret_r=1)

        CS2 = CalibrationSequence(np.array([0, 45]), np.array([90, 45]),
                                  np.array([True, True]), np.array([True, True]), np.array([False, False]),
                                  np.array([1, 2]), ret_0_h=1, dret_h=1, ret_0_45=1, dret_45=2, ret_0_r=0, dret_r=1)
        CS = CS1 + CS2

        truth1 = Data.rotation(-np.pi/2) @ Data.elliptical_retarder(1, 1, 1, 0) @ Data.rotation(np.pi / 2.) @ Data.polarizer(1, 0)
        truth2 = np.diag(np.ones(4))
        truth3 = Data.rotation(-np.pi/2) @ Data.elliptical_retarder(1, 1, 1, 0) @ Data.rotation(np.pi / 2.) @ Data.polarizer(1, 0)
        truth4 = Data.rotation(-np.pi/4) @ Data.elliptical_retarder(1, 2, 3, 1) @ Data.rotation(np.pi / 4.) @ Data.rotation(-np.pi / 4.) @ Data.polarizer(1, 0) @ Data.rotation(np.pi / 4)

        self.assertEqual(CS.CM.shape, (4,4,4))
        np.testing.assert_allclose(CS.CM[0], truth1)
        np.testing.assert_allclose(CS.CM[1], truth2)
        np.testing.assert_allclose(CS.CM[2], truth3)
        np.testing.assert_allclose(CS.CM[3], truth4)



import yaml
import pkg_resources

s122_file = pkg_resources.resource_filename('dkist_processing_pac', 'data/SPEC_0122_mapping.yml')
with open(s122_file, 'rb') as f:
    S122 = yaml.safe_load(f)

constants_file = pkg_resources.resource_filename('dkist_processing_pac', 'data/constants.yml')
with open(constants_file, 'rb') as f:
    CONSTANTS = yaml.safe_load(f)

def load_visp_keywords() -> None:
    """Load the ViSP Instrument keywords from the ViSP Pipeline into the S122 dictionary
    """

    from ViSP_Pipeline.data import S122 as ViSP_S122

    PAC_keys = list(S122.keys())  # I think this is faster than generating it for every visp key?
    for k in ViSP_S122.keys():
        if k not in PAC_keys:
            S122['ViSP_{}'.format(k)] = ViSP_S122[k]

def load_cryo_keywords() -> None:
    """Load the CryoNIRSP Instrument keywords into the S122 dictionary
    """

    from CryoNIRSP_Pipeline.data import S122 as Cryo_S122

    PAC_keys = list(S122.keys())
    for k in Cryo_S122.keys():
        if k not in PAC_keys:
            S122['CRSP_{}'.format(k)] = Cryo_S122[k]

def load_dlnirsp_keywords() -> None:
    """Load the DLNIRSP Instrument keywords into the S122 dictionary
    """

    from DLNIRSP_Pipeline.data import S122 as DL_S122

    PAC_keys = list(S122.keys())
    for k in DL_S122.keys():
        if k not in PAC_keys:
            S122['DLN_{}'.format(k)] = DL_S122[k]



import hashlib
import numpy as np
from dkist_processing_pac.data import CONSTANTS

def compute_X12(i2q, i2u):
    """Compute the X12 mirror parameter from measures of continuum polarization

    For now this assumes no rotation b/t M2 and the CU. This might change in the future.

    Parameters
    ----------
    i2q : float
        Q/I in a continuum region

    i2u : float
        U/I in a continuum region

    Returns
    -------
    float
        X12 mirror parameter

    """
    X12 = np.sqrt((1 - i2q) / (1 + i2q))

    return X12

def update_xtalk_database(i2q, i2u, instrument_name, date, wave, notes='', db_file=None):
    """Update the monitoring database that contains every measurement of X12 from all science programs

    Each entry is given a unique hash in addition to keeping track of instrument, date, and wavelength. This ensures
    that potential race-condition collisions don't happen.

    Parameters
    ----------
    i2q : float
        Q/I in a continuum region

    i2u : float
        Q/U in a continuum region

    instrument_name : str
        The name of the instrument providing this information. Truncated to 20 characters.

    date : float
        The current time/time of observations in MJD

    wave : float
        The wavelength of observations in nm

    notes : str
        Any user notes. Spaces will be converted to _

    db_file : str
        If provided, the path to the X12 database file. If None then the default file will be loaded.

    Returns
    -------
    None
    """
    X12 = compute_X12(i2q, i2u)

    info_str = '{}{}{}{}{}'.format(instrument_name, date, i2q, i2u, X12)
    hash_str = hashlib.sha256(info_str.encode()).hexdigest()

    if db_file is None:
        db_file = CONSTANTS['xtalk_db_file']

    with open(db_file, 'a') as f:
        f.write('{:66}{:20.20} {:20.20} {:20.10f} {:7.2f} {:10.6f} {:10.6f} {:10.6f}\n'.format(hash_str,
                                                                                               instrument_name.replace(' ','_'),
                                                                                               notes.replace(' ','_'),
                                                                                               date, wave,
                                                                                               i2q, i2u, X12))

    return

def get_updated_X12(db_file=None):
    """Return the 'best' X12 value from the monitoring database.

    Right now this just returns the most recent value. This will probably change at some point.

    Parameters
    ----------
    db_file : str
        If provided, the path to the X12 database file. If None then the default file will be loaded.

    Returns
    -------
    float
        The most up-to-date value of X12
    """
    if db_file is None:
        db_file = CONSTANTS['xtalk_db_file']

    mjd, wave, X12 = np.loadtxt(db_file, usecols=(3, 4, 7), unpack=True)

    sidx = np.argsort(mjd)

    return X12[sidx[-1]]

def write_blank_db_file(db_file=None):
    """Generate a blank X12 monitoring file.

    Really this just writes the header

    Parameters
    ----------
    db_file : str
        If provided, the path to the X12 database file. If None then the default file will be loaded.

    Returns
    -------
    None
    """
    if db_file is None:
        db_file = CONSTANTS['xtalk_db_file']

    with open(db_file, 'w') as f:
        f.write('# {:64}{:20} {:20} {:>20} {:>7} {:>10} {:>10} {:>10}\n\n'.format('Unique hash',
                                                                                 'Instrument Name',
                                                                                 'Notes', 'Date [MJD]', 'Wave',
                                                                                 'Q/I','Q/U','X12'))

    return



import os
import numpy as np
from astropy.io import fits as pyfits
from dkist_processing_pac import Data, CUModel, TelescopeModel, FittingFramework, generic
from dkist_processing_pac.tag import tag


def main(data_dir, CU_par_file, dmod_output, overwrite=True, suffix='.FITS', telescope_db='telescope_db.txt',
         skip_darks=True):
    """Load previously-computed CU model parameters and use them to compute demodulation matrices.

    This function also requires a Drawer of PolCal data, which should be the exact same data used to fit the CU
    parameters. It is OK (and expected) that these data and the CU parameters themselves will be resampled from the
    field-sampling used to fit the CU parameters. This allows the Instrument Partners to define different Regions of
    Calibration Unit Homogeneity for the fit of parameters and the calculation of demodulation matrices.

    If the input PolCal data have shape (X, Y, M, N) where (X, Y) are the field shape of the data, M is the number
    of modulation states, and N is the number of CS steps then the resulting demodulation matrices will have shape
    (X, Y, 4, M) and will be saved to the first ImageHDU of a single FITS file. The header for this FITS file is a copy
    of the header in the Drawer's first CS step.

    Parameters
    ----------
    data_dir: str
        Location of the directory containing reduced data from a single PolCal sequence (i.e., a Drawer)

    CU_par_file : str
        Location of file containing a set of CU model parameters at each (x, y) field location in the Drawer

    dmod_output : str
        Location in which to save the final demodulation matrices

    overwrite : bool
        If True then ``dmod_output`` will be overwritten if it exists

    suffix : str
        Only files in ``data_dir`` with this suffix will be loaded

    telescope_db : str
        Location of the file containing Telescope Model parameters as a function of time and wavelength

    skip_darks : bool
        If True (default) then don't include and dark steps of the CS in fitting

    """
    if os.path.exists(dmod_output) and not overwrite:
        raise IOError('File {} already exists and overwrite is False'.format(dmod_output))

    DRWR = Data.Drawer(data_dir, suffix=suffix, skip_darks=skip_darks)
    print('{}: loaded SoCC from {}'.format(tag(), data_dir))

    CMP = Data.CUModelParams(*(DRWR.shape + (1, DRWR.nummod)))
    CMP.loadfits(CU_par_file)
    print('{}: loaded CU Model Parameters from {}'.format(tag(), CU_par_file))

    demodulation = core_demod(DRWR, CMP, telescope_db)

    print('{}: saving demodulation matrices to {}'.format(tag(), dmod_output))
    primary = pyfits.PrimaryHDU(header=DRWR.data_list[0][0].header)
    primary.header['OBJECT'] = 'Demodulation matrix'

    dmod_hdu = pyfits.ImageHDU(demodulation, header=DRWR.data_list[0][0].header)
    dmod_hdu.header['OBJECT'] = 'Demodulation matrix'
    pyfits.HDUList([primary, dmod_hdu]).writeto(dmod_output, overwrite=overwrite)
    del DRWR
    CMP.cleanup()

    return

def DC_main(dresser: Data.Dresser, cu_params: Data.CUModelParams, telescope_db: str) -> np.ndarray:

    if len(dresser.drawers) == 0:
        raise ValueError('The provided dresser has no Drawers!')
    if len(dresser.drawers) > 1:
        raise ValueError('The provided dresser has more than one Drawer')

    drawer = dresser.drawers[0]

    return core_demod(drawer, cu_params, telescope_db)

def core_demod(drawer: Data.Drawer, cu_params: Data.CUModelParams, telescope_db: str) -> np.ndarray:

    mode_opts = generic.init_fit_mode(cu_params.fit_mode, drawer.wavelength, False, cu_params.init_set)
    use_T = mode_opts['switches']['use_T']
    use_M12 = mode_opts['switches']['use_M12']

    CS = CUModel.CalibrationSequence(drawer.theta_pol_steps, drawer.theta_ret_steps, drawer.pol_in, drawer.ret_in,
                                     drawer.dark_in, drawer.timeobs)
    TM = TelescopeModel.TelescopeModel(drawer.azimuth, drawer.elevation, drawer.table_angle)

    print('{}: updating polarizer py value'.format(tag()))
    wave = drawer.wavelength
    CS.set_py_from_database(wave)

    print('{}: loading telescope database from {}'.format(tag(), telescope_db))
    mean_time = 0.5 * (drawer.date_bgn + drawer.date_end)
    TM.load_from_database(telescope_db, mean_time, wave)
    numx, numy, numl = drawer.shape
    nummod = drawer.nummod
    demodulation = np.zeros((numx, numy, numl, 4, nummod), dtype=np.float64)

    print('{}: shape of Drawer: ({}, {}, {})'.format(tag(), numx, numy, numl))
    print('{}: computing demoduation matrices'.format(tag()))
    for i in range(numx):
        for j in range(numy):
            for k in range(numl):
                print('{}:  position = ({}, {}, {})'.format(tag(), i, j, k))
                I_cal = drawer[i, j, k]
                CS.load_pars_from_dict(cu_params[i, j, k])

                print('{}:  fitting modulation matrix to best CU model'.format(tag()))
                S = FittingFramework.generate_S(TM, CS, use_T=use_T, use_M12=use_M12)
                try:
                    O = FittingFramework.fit_modulation_matrix(I_cal, S)
                except ValueError as e:
                    if 'illegal value' in str(e):
                        print('{}:  Garbage results detected. Setting pass-through demodulation matrix.'.format(tag()))
                        O = np.zeros((nummod, 4))
                        O[:, 0] = 1./nummod
                    else:
                        raise

                print('{}:  computing inverse of modulation matrix'.format(tag()))
                demodulation[i, j, k, :, :] = np.linalg.pinv(O)

    return demodulation

def command_line():
    import argparse
    import sys

    parser = argparse.ArgumentParser(description='Generate demodulation matrices from a single Calibration Sequence. '
                                     'The data directory specified must contain a separate FITS file for each step in '
                                     'the CS that has already been processed by the instrument reduction code. The '
                                     'CU par file is the output of `pac_cu_par`.')
    parser.add_argument('data_dir', help='Directory containing a single Calibration Sequence')
    parser.add_argument('CU_pars', help='FITS file containing best fit Calibration Unit parameters')
    parser.add_argument('output', help='Location of FITS file to save demodulation matrices')
    parser.add_argument('-f', '--force', action='store_true', help='Overwrite outputs')
    parser.add_argument('-t', '--telescope-db', help='Database listing telescope parameters as function of obstime and '
                                                     'wavelength. This option is ignored if telescope_db is defined in '
                                                     'a provided CONFIG_FILE (-c).', default=None)
    parser.add_argument('-s', '--suffix', help='File suffix to filter data. File mask is data_dir/*suffix',
                        default='.FITS')
    parser.add_argument('-D', '--use-darks', action='store_false', help='Include any DARK CS steps in the fit')
    args = parser.parse_args()

    if args.telescope_db is None:
        args.telescope_db = generic.get_default_telescope_db()

    if os.path.exists(args.data_dir):
        sys.exit(main(args.data_dir, args.CU_pars, args.output, suffix=args.suffix, overwrite=args.force,
                      telescope_db=args.telescope_db, skip_darks=args.use_darks))
    else:
        print('Directory {} does not exist. Aborting.'.format(args.data_dir))
        sys.exit(1)




import corner
import pickle
import time
import re
import os
import numpy as np
from glob import glob
from astropy.io import fits as pyfits
from dkist_processing_pac import FittingFramework, CUModel, Data, TelescopeModel
from dkist_processing_pac.tag import tag
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages as PDF

def main_MCMC(aux_dir, **mckwargs):
    """Read in a directory of PA&C fits and run MCMC analysis to explore the fit parameter space.

    Parameters
    ----------
    aux_dir : str
        Path to a directory containing the "auxiliary" fitting information. This is usually generated automatically by
        `pac_demod` or `pac_tele`.

    outpre : str
        Prefix to use when saving the results of MCMC analysis

    mckwargs : dict
        A dictionary of options to pass to the MCMC algorithm.

    Returns
    -------
    None
    """
    pickle_list = glob('{}/fitobj*.pkl'.format(aux_dir))
    print('{}: found fits for {} detector positions'.format(tag(), len(pickle_list)))

    for p in pickle_list:
        xpos, ypos, lpos = re.search('_(\d{4})(\d{4})(\d{4})\.pkl', p).groups()
        print('{}: analyzing position ({}, {}, {}) from {}'.format(tag(), int(xpos), int(ypos), int(lpos), p))
        analyze_pickle(p, '{}/mcmcsav_{}{}{}'.format(aux_dir, xpos, ypos, lpos), **mckwargs)

    return

def analyze_pickle(pickle_file, outpre, plot_burn=0, **mckwargs):
    """Run MCMC analysis on a single PA&C fit object

    An MCMC algorithm is first used to explore the parameter space of all free variables found in the fit object.
    Corner and Trace plots are then produced to aid the user in interpreting the MCMC results.

    Parameters
    ----------
    pickle_file : str
        Path to the pickel (*.pkl) file containing `lmfit` fit objects. This is probably generated automatically by
        `pac_demod` or `pac_tele`.

    outpre : str
        Prefix to use when saving MCMC results.

    plot_burn : int
        Number of MCMC samples to burn ONLY WHEN PLOTTING. This is useful if no samples were burned during the actual
        MCMC run and you want to test different burn amounts when analyzing the results.

    mckwargs : dict
        Additional arguments to pass to the MCMC algorithm.

    Returns
    -------
    lmfit.Minimizer
        The original fitting object from the base PA&C fits

    lmfit.MinimizerResult
        The original fit result object from the vase PA&C fits

    lmfit.MinimizerResult
        The result object that contains the results of the MCMC run

    """
    with open(pickle_file, 'rb') as f:
        fitter = pickle.load(f)
        mini_res = pickle.load(f)

    expand_param_range(mini_res, factor=100)
    print('{}:  starting MCMC analysis with keywords {}'.format(tag(), mckwargs))
    mcmc_res = run_MCMC(fitter, mini_res, **mckwargs)

    print('{}:  plotting results'.format(tag()))
    plot_MCMC(mcmc_res, mini_res, outpre, plot_burn=plot_burn)

    pkl_save_name = outpre + '.pkl'
    print('{}:  saving results to {}'.format(tag(), pkl_save_name))
    with open(pkl_save_name, 'wb') as f:
        pickle.dump(fitter, f)
        pickle.dump(mini_res, f)
        pickle.dump(mcmc_res, f)

    return fitter, mini_res, mcmc_res

def run_MCMC(fitter, minimize_results, burn=0, steps=100, nwalkers=256, **mckwargs):
    """Initialize and run MCMC analysis on fit PA&C fit object

    The `emcee` package is used to perform the MCMC analysis

    Parameters
    ----------
    fitter : lmfit.Minimizer
        Fit object of the original PA&C parameter fits

    minimize_results : lmfit.MinimizerResult
        Fitting results of the original PA&C parameter fits

    burn : int
        The number of MCMC steps to discard from the beginning of the Markov Chain. This is used to allow the walkers
        to settle into a suitably stochastic state.

    steps : int
        The number of steps for each MCMC walker to take

    nwalkers : int
        The number of MCMC walkers to use

    mckwargs : dict
        Any additional parameters to pass to `emcee`

    Returns
    -------
    lmfit.MinimizerResult
        Fit result object containing the results of MCMC analysis. Most usefully, the MCMC chain.

    """
    FittingFramework.print_header(minimize_results.params)
    t1 = time.time()
    fitter.userargs = (*fitter.userargs[:-1], t1)
    mcmc_results = fitter.emcee(burn=burn, steps=steps, nwalkers=nwalkers, params=minimize_results.params, **mckwargs)
    print('\n{}: completed in {:4.3f} s'.format(tag(), time.time() - t1))

    return mcmc_results

def expand_param_range(fit_results, params='all', factor=30):
    """Expand the original PA&C fit parameter ranges to allow for a broader search by the MCMC walkers

    This is a very good idea to do because if the MCMC walkers are prevented from exploring the entire distribution of
    each parameter then the resulting conclusions about credibility regions will be invalid

    Parameters
    ----------
    fit_results : lmfit.MinimizerResult
        Fitting results of the original PA&C parameter fits

    params : str or list
        The name or list of names of parameters to expand. The given parameters must exist in fit_results.var_names.
        If set to 'all' then all free parameters that are found will be expanded.

    factor : float
        The percentage by which to expand the parameters. The minimum and maximum will be shifted by factor% of their
        original values.
    """
    if params == 'all':
        params = fit_results.var_names
    elif type(params) is not list:
        params = [params]

    for p in params:
        fit_results.params[p].min *= 1 - factor / 100 * np.sign(fit_results.params[p].min)
        fit_results.params[p].max *= 1 + factor / 100 * np.sign(fit_results.params[p].max)

def plot_MCMC(mcres, minimize_results, outpre, plot_burn=0):
    """Plot corner and trace plots for an MCMC run and save them to disk.

    The plots also report the credibility region for each parameter and show the compare the mean and median of the MCMC
    samples to the original fit value.

    Parameters
    ----------
    mcres : lmfit.MinimizerResults
        Results from an lmfit MCMC run

    minimize_results : lmfit.MinimizerResults
        Results from original PA&C parameter fitting

    outpre : str
        Prefix for output files. The files are OUTPRE_corner.pdf and OUTPRE_traces.pdf

    plot_burn : int
        The number of MCMC samples to exclude at the start of the chain

    Returns
    -------
    numpy.ndarray
        The best-fit parameter values from the original PA&C fit

    numpy.ndarray
        The median parameter values from the MCMC run
    """
    labels = texify_labels(mcres)

    minimizer_truths = np.array([minimize_results.params.valuesdict()[p] for p in minimize_results.var_names])
    mc_meds = np.median(mcres.flatchain, axis=0)
    mc_means = np.mean(mcres.flatchain, axis=0)
    if '__lnsigma' in mcres.var_names:
        minimizer_truths = np.append(minimizer_truths, mc_meds[-1])
    flatchain = mcres.chain[:, plot_burn:, :].reshape(-1, mcres.chain.shape[-1])
    cfig = corner.corner(flatchain, labels=labels,
                         show_titles=True, title_kwargs={'fontsize': 12})

    truth_colors = ['#1b9e77','#d95f02','#7570b3']
    numvars = mcres.chain.shape[-1]
    axes = np.array(cfig.axes).reshape((numvars, numvars))
    for i in range(numvars):
        ax = axes[i, i]
        ax.axvline(minimizer_truths[i], color=truth_colors[0])
        ax.axvline(mc_meds[i], color=truth_colors[1])
        ax.axvline(mc_means[i], color=truth_colors[2])

    for yi in range(numvars):
        for xi in range(yi):
            ax = axes[yi, xi]
            ax.axvline(minimizer_truths[xi], color=truth_colors[0])
            ax.axvline(mc_meds[xi], color=truth_colors[1])
            ax.axvline(mc_means[xi], color=truth_colors[2])
            ax.axhline(minimizer_truths[yi], color=truth_colors[0])
            ax.axhline(mc_meds[yi], color=truth_colors[1])
            ax.axhline(mc_means[yi], color=truth_colors[2])
            ax.plot(minimizer_truths[xi], minimizer_truths[yi], 's', color=truth_colors[0])
            ax.plot(mc_meds[xi], mc_meds[yi], 's', color=truth_colors[1])
            ax.plot(mc_means[xi], mc_means[yi], 's', color=truth_colors[2])

    axes[0, 1].text(0,0.2, 'minimzer value', color=truth_colors[0], transform=axes[0, 1].transAxes, fontsize=14)
    axes[0, 1].text(0,0.3, 'MCMC median', color=truth_colors[1], transform=axes[0, 1].transAxes, fontsize=14)
    axes[0, 1].text(0,0.4, 'MCMC mean', color=truth_colors[2], transform=axes[0, 1].transAxes, fontsize=14)

    cfig.savefig(outpre + '_corner.pdf')

    pp = PDF(outpre + '_traces.pdf')
    for i in range(numvars):
        if i % 3 == 0:
            fig = plt.figure()
        ax = fig.add_subplot(3,1,(i % 3) + 1)
        for j in range(mcres.chain.shape[0]):
            ax.plot(mcres.chain[j,plot_burn:,i], color='k', alpha=0.2, lw=0.6)
        ax.set_ylabel(labels[i])
        ax.axhline(minimizer_truths[i], color=truth_colors[0], ls='-', lw=0.6)
        ax.axhline(mc_meds[i], color=truth_colors[1], ls='-', lw=0.6)
        ax.axhline(mc_means[i], color=truth_colors[2], ls='-', lw=0.6)
        if i % 3 == 2:
            fig.subplots_adjust(hspace=0.0001)
            pp.savefig(fig)

    # If we ended on a non-full page we need to save it to the PDF
    if i % 3 != 2:
        fig.subplots_adjust(hspace=0.0001)
        pp.savefig(fig)

    pp.close()
    plt.close('all')

    return minimizer_truths, mc_meds


def texify_labels(mcres):
    """Convert the variable names used by lmfit to nice-looking latex labels for plotting

    Only parameters that were varried during the fit are converted.

    This function uses regular expressions for no reason at all and is perhaps too complicated.

    Parameters
    ----------
    mcres : lmfit.MinimzerResults
        Results from an lmfit fitting run

    Returns
    -------
    list
        List containing latex-style variable names
    """

    labels = []
    for l in mcres.var_names:
        if l == '__lnsigma':
            labels.append(r'$\mathrm{ln}\sigma$')
            continue
        try:
            tmp = re.sub('^ret0(.*)_', r'\\delta_{0,\1}_', l)
            tmp = re.sub('^dret(.*)_', r'\\Delta\\delta_{\1}_', tmp)
            tmp = re.sub('^theta(.*)_', r'\\theta_{\1}_', tmp)
            tmp = re.sub(r'_CS(\d\d)', r'^{\1}', tmp)
            tmp = re.sub(r'_(\w+)', r'_{\1}', tmp)
            tmp = re.sub(r'\^{0(\d)}', r'^{\1}', tmp)
            tmp = re.sub(r'^t(\d\d)', r'\\tau_{\1}', tmp)
            tmp = re.sub(r'^x(\d\d)', r'X_{\1}', tmp)
            tmp = r'$' + tmp + r'$'
        except:
            labels.append(l)

        labels.append(tmp)

    return labels

def main_demod_err(data_dir, demod_file, output, suffix='*.FITS', telescope_db='telescope_db.txt', overwrite=True,
                   skip_darks=True, plot=True):
    """Create uncertainty estimates for a given demodulation matrix using the results of MCMC analysis.

    The result is the uncertainty analog of the output of `pac_demod` along with some optional diagnostic plots. Just
    like the main demodulation matrices, the uncertainty estimates will be contained in the 1st ImageHDU of a FITS file
    in an array of shape (X, Y, 4, M), even if X = Y = 1.

    Parameters
    ----------
    data_dir : str
        Path to directory containing data related to a single Calibration Sequence

    demod_file : str
        Path to the demodulation file produced by running `pac_demod` on the data in data_dir

    output : str
        Path to output file

    suffix : str
        File suffix to use when loading the Calibration Sequence data from data_dir. The file mask is DATA_DIR/*SUFFIX

    telescope_db : str
        Path to database of telescope parameters as a function of time and wavelength

    overwrite : bool
        If True then output will be overwritten if it exists

    plot : bool
        If True then a plots showing the distribution of values for each entry of the demodulation matrix will be
         created. Each of the (X, Y) positions gets its own plot and they are all saved to OUTPUT.pdf

    skip_darks : bool
        If True (default) then don't include and dark steps of the CS in fitting

    Returns
    -------

    """
    dirname = os.path.dirname(demod_file) or '.'
    pickle_path = '{}/fitaux'.format(dirname)

    try:
        tmpkl = glob('{}/mcmcsav*.pkl'.format(pickle_path))[0]
    except IndexError:
        raise FileNotFoundError('Cannot find any MCMC results in {}'.format(dirname))

    with open(tmpkl, 'rb') as f:
        fitter = pickle.load(f)
    linear = 'b_CS00' in fitter.params.keys()

    PCD = Data.Drawer(data_dir, suffix, skip_darks=skip_darks)
    TM = TelescopeModel.TelescopeModel(PCD.azimuth, PCD.elevation, PCD.table_angle)
    if linear:
        from dkist_processing_pac.linear_retarder import CULinearRetarderModel
        CS = CULinearRetarderModel.CalibrationSequence(PCD.theta_pol_steps, PCD.theta_ret_steps,
                                                       PCD.pol_in, PCD.ret_in, PCD.timeobs)
    else:
        CS = CUModel.CalibrationSequence(PCD.theta_pol_steps, PCD.theta_ret_steps, PCD.pol_in, PCD.ret_in,
                                         PCD.dark_in, PCD.timeobs)
    print('{}: loaded polcal data from {}'.format(tag(), data_dir))

    print('{}: updating polarizer py value'.format(tag()))
    wave = PCD.wavelength
    CS.set_py_from_database(wave)

    print('{}: loading telescope database from {}'.format(tag(), telescope_db))
    mean_time = 0.5 * (PCD.date_bgn + PCD.date_end)
    TM.load_from_database(telescope_db, mean_time, wave)

    numx, numy, numl = PCD.shape
    demod_err = np.zeros((numx, numy, numl, 4, PCD.nummod), dtype=np.float32)

    pickle_list = glob('{}/mcmcsav*.pkl'.format(pickle_path))
    print('{}: found fits for {} detector positions'.format(tag(), len(pickle_list)))

    if plot:
        plot_prefix = '.'.join(output.split('.')[:-1]) or output
        plotfile = '{}.pdf'.format(plot_prefix)
        pp = PDF(plotfile)

    for i in range(numx):
        for j in range(numy):
            for l in range(numl):
                print('{}: computing err for position ({}, {}, {})'.format(tag(), i, j, l))

                I = PCD[i, j, l]
                pickle_file = '{}/mcmcsav_{:04}{:04}{:04}.pkl'.format(pickle_path, i, j, l)
                if not os.path.exists(pickle_file):
                    raise FileNotFoundError('Cannot find MCMC results {} for position ({}, {})'.format(pickle_file, i, j))

                if linear:
                    demod_stack = get_linear_demod_stack(pickle_file, I, TM, CS)
                else:
                    demod_stack = get_demod_stack(pickle_file, I, TM, CS)

                if plot:
                    fig = plot_demod_err(demod_stack)
                    fig.suptitle('Position ({}, {}) from {}'.format(i, j, pickle_file))
                    pp.savefig(fig)
                    plt.close(fig)

                demod_err[i, j, l] = np.std(demod_stack, axis=0)

    demod_hdus = pyfits.open(demod_file)
    primary = pyfits.PrimaryHDU(header=demod_hdus[0].header)
    primary.header['OBJECT'] = 'Demodulation matrix uncertainty'

    demod_hdu = pyfits.ImageHDU(demod_err, header=demod_hdus[1].header)
    demod_hdu.header['OBJECT'] = 'Demodulation matrix uncertainty'
    pyfits.HDUList([primary, demod_hdu]).writeto(output, overwrite=overwrite)

    if plot:
        print('{}: saving plots to {}'.format(tag(), plotfile))
        pp.close()

    return demod_err, demod_stack

#TODO: Update this to be more flexible with fit parameters
def get_demod_stack(pickle_file, I, TM, CS, numload=None):
    """Compute a demodulation matrix for every MCMC step represented in a MCMC results file.

    Demodulation matrices are computed with `FitModels.fit_modulation_matrix`.

    Parameters
    ----------
    pickle_file : str
        Location of MCMC results pickle file

    I : numpy.ndarray
        Array of shape (M, N) containing the observed intensity values where M is the number of modulation states and
        N is the number of CS steps

    TM : TelescopeModel.TelescopeModel
        An object describing the telescope configuration at each step in the CS

    CS : CUModel.CalibrationSequence
        An object describing the CU configuration at each step in the CS

    numload : int
        The maximum number of MCMC steps to load. Useful for debuggin.

    Returns
    -------
    numpy.ndarray
        Array of shape (Z, 4, M) where Z is the min(numload, number of MCMC steps). Each of the Z arrays is a single
        demodulation matrix.
    """
    with open(pickle_file, 'rb') as f:
        fitter = pickle.load(f)
        results = pickle.load(f)
        mcres = pickle.load(f)
    flatchain = mcres.flatchain
    numsteps = flatchain.shape[0]
    print('{}: loaded MCMC results from {}'.format(tag(), pickle_file))

    demod_list = []
    for k in range(numsteps)[:numload]:
        print('\r{}: reading MCMC steps: {}/{} ({}%)'.format(tag(), k + 1, numsteps,
                                                             int((k + 1) / numsteps * 100)), end='', flush=True)
        CS.ret_0_h[0] = flatchain['ret0h_CS00'][k]
        CS.dret_h[0] = flatchain['dreth_CS00'][k]
        CS.ret_0_45[0] = flatchain['ret045_CS00'][k]
        CS.dret_45[0] = flatchain['dret45_CS00'][k]
        CS.ret_0_r[0] = flatchain['ret0r_CS00'][k]
        CS.dret_r[0] = flatchain['dretr_CS00'][k]

        S = (TM.TM @ CS.CM @ TM.M12 @ CS.s_in).T

        O = FittingFramework.fit_modulation_matrix(I, S)
        demod_list.append(np.linalg.pinv(O))

    print('')
    demod_stack = np.stack(demod_list, 0)

    return demod_stack

def get_linear_demod_stack(pickle_file, I, TM, CS, numload=None):
    """Compute a demodulation matrix for every MCMC step represented in a MCMC results file.

    This function expects the fits to been done with a CU linear retarder model.

    Demodulation matrices are computed with `FitModels.fit_modulation_matrix`.

    Parameters
    ----------
    pickle_file : str
        Location of MCMC results pickle file

    I : numpy.ndarray
        Array of shape (M, N) containing the observed intensity values where M is the number of modulation states and
        N is the number of CS steps

    TM : TelescopeModel.TelescopeModel
        An object describing the telescope configuration at each step in the CS

    CS : CULinearRetarderModel.CalibrationSequence
        An object describing the CU configuration at each step in the CS

    numload : int
        The maximum number of MCMC steps to load. Useful for debuggin.

    Returns
    -------
    numpy.ndarray
        Array of shape (Z, 4, M) where Z is the min(numload, number of MCMC steps). Each of the Z arrays is a single
        demodulation matrix.
    """
    with open(pickle_file, 'rb') as f:
        fitter = pickle.load(f)
        results = pickle.load(f)
        mcres = pickle.load(f)
    flatchain = mcres.flatchain
    numsteps = flatchain.shape[0]
    print('{}: loaded MCMC results from {}'.format(tag(), pickle_file))

    demod_list = []
    for k in range(numsteps)[:numload]:
        print('\r{}: reading MCMC steps: {}/{} ({}%)'.format(tag(), k + 1, numsteps,
                                                             int((k + 1) / numsteps * 100)), end='', flush=True)
        CS.ret_0[0] = flatchain['ret0_CS00'][k]
        CS.dret[0] = flatchain['dret_CS00'][k]
        CS.theta_off[0] = flatchain['theta_off_CS00'][k]
        try:
            CS.diatten[0] = flatchain['b_CS00'][k]
        except KeyError:
            pass

        S = (TM.TM @ CS.CM @ TM.M12 @ CS.s_in).T

        O = FittingFramework.fit_modulation_matrix(I, S)
        demod_list.append(np.linalg.pinv(O))

    print('')
    demod_stack = np.stack(demod_list, 0)

    return demod_stack

def plot_demod_err(demod_stack, output=None):
    """Make a plot showing the distributions of values for each entry in a demodulation matrix

    The stddev and stddev/mean will be printed in each cell

    Parameters
    ----------
    demod_stack : numpy.ndarray
        Array of shape (Z, 4, M) where Z is the min(numload, number of MCMC steps). Each of the Z arrays is a single
        demodulation matrix. Probably the output of `get_demod_stack`.

    output : str
        If True then save the generated plot to the specified file.

    Returns
    -------
    matplotlib.figure.Figure
        Figure object containing the grid of plots that represents the distribution of the demodulation matrix
    """
    print('{}: generating demodulation uncertainty plot'.format(tag()))

    nummod = demod_stack.shape[2]

    std = np.std(demod_stack, axis=0)
    mean = np.mean(demod_stack, axis=0)

    fig = plt.figure(figsize=(16,22))
    axes = fig.subplots(nummod,4,gridspec_kw={'wspace': 0.02,
                                              'left': 0.05, 'right': 0.95, 'bottom': 0.03, 'top':0.95})

    lims = []
    for i in range(nummod):
        for j in range(4):
            ax = axes[i, j]
            ax.hist(demod_stack[:,j,i], histtype='stepfilled')
            ax.set_yticks([])
            ax.text(0.98, 0.9, r'$\sigma_D = {:5.2e}$'.format(std[j, i]), transform=ax.transAxes,
                    horizontalalignment='right')
            ax.text(0.98, 0.8, r'$\frac{{\sigma_D}}{{\left<D\right>}} = {:5.2e}$'.format(std[j, i]/mean[j, i]),
                    transform=ax.transAxes, horizontalalignment='right')
            lims.append(ax.get_xlim())

    half_width = 0.5 * np.max(np.diff(np.array(lims), axis=1))
    for ax in fig.axes:
        mid = np.mean(ax.get_xlim())
        ax.set_xlim(mid - half_width, mid + half_width)

    fig.subplots_adjust(wspace=0.01)

    if output is not None:
        fig.savefig(output)

    return fig

def command_line_MCMC():
    import argparse
    import sys

    parser = argparse.ArgumentParser(description='Compute PA&C fit uncertainties using Markov-Chain Monte Carlo (MCMC) '
                                                 'analysis. This program operates on the auxiliary output of '
                                                 '"pac_demod" or "pac_tele", which is generated with the "-S" option.')
    parser.add_argument('aux_dir', help='Directory containing auxiliary fit information')
    parser.add_argument('-b', '--burn', nargs=1, default=[0], type=int, help='Number of MCMC steps to burn')
    parser.add_argument('-N', '--num-steps', nargs=1, default=[100], type=int,
                        help='Number of MCMC steps for each walker to take')
    parser.add_argument('-w', '--num-walkers', nargs=1, default=[256], type=int, help='Number of MCMC walkers to use')
    args = parser.parse_args()

    if os.path.exists(args.aux_dir):
        sys.exit(main_MCMC(args.aux_dir, burn=args.burn[0], nwalkers=args.num_walkers[0], steps=args.num_steps[0]))
    else:
        print('Directory {} does not exist. Aborting.'.format(args.aux))
        sys.exit(1)

def command_line_demod_err():
    import argparse
    import sys
    from dkist_processing_pac import generic

    parser = argparse.ArgumentParser(description='Generate demodulation matrix uncertainties using the results of '
                                                 'Markov-Chain Monte Carlo (MCMC) analysis of CU parameter fits. This '
                                                 'program operates on the output of "pac_mcmc".')
    parser.add_argument('data_dir', help='Directory containing a single Calibration Sequence. This is the same '
                                         'directory that was the input to "pac_demod".')
    parser.add_argument('demod_file', help='The Demodulation matrix file generated by "pac_demod"')
    parser.add_argument('output', help='Location of FITS file to save uncertainty on Demodulation matrices')
    parser.add_argument('-s', '--suffix', help='File suffix to filter data. File mask is data_dir/*suffix',
                        default='.FITS')
    parser.add_argument('-t', '--telescope-db', help='Location of database containing telescope parameters as a '
                                                     'function of date and wavelength', nargs=1, default=[])
    parser.add_argument('-p', '--plot', action='store_true', help='Generate plots showing the value distribution of '
                                                                  'each entry in the Demodulation matrices')
    parser.add_argument('-D', '--use-darks', action='store_false', help='Include any DARK CS steps in the fit')
    args = parser.parse_args()

    if not args.telescope_db:
        telescope_db = generic.get_default_telescope_db()
    else:
        telescope_db = args.telescope_db[0]

    if not os.path.exists(args.data_dir):
        print('Directory {} does not exist. Aborting.'.format(args.data_dir))
        sys.exit(1)

    if not os.path.exists(args.demod_file):
        print('Demodulation matrices file {} does not exist. Aborting.'.format(args.demod_file))
        sys.exit(1)

    main_demod_err(args.data_dir, args.demod_file, args.output, suffix=args.suffix,
                   telescope_db=telescope_db, plot=args.plot, skip_darks=args.use_darks)




from unittest import TestCase

from dkist_processing_pac import generic

#TODO: Make this whole module not stupid. It has become stupid since the switch to ini_params.asdf implementation
class TestFitRecipes(TestCase):

    def setUp(self):
        self.exclude = ['x34','t34','x56','t56','switches','I_sys']

    def param_in_range(self, mode_opts, parameter):
        assert mode_opts[parameter]['value'] >= mode_opts[parameter][ 'min']
        assert mode_opts[parameter]['value'] <= mode_opts[parameter][ 'max']

    def test_baseline(self):
        mode_opts = generic.init_fit_mode('baseline', 666, True)
        for par in mode_opts.keys():
            if par in self.exclude:
                continue
            with self.subTest(parameter=par):
                self.param_in_range(mode_opts, par)



import os
import pytest
import pkg_resources
from unittest import TestCase
import numpy as np
from astropy.time import Time
from dkist_processing_pac import Data
from dkist_processing_pac.utils import gen_fake_data

RTOL=1e-6
test_dir = pkg_resources.resource_filename('dkist_processing_pac', 'tests/data')
RUN1=test_dir+'/run1'
RUN2=test_dir+'/run2'
BADINST=test_dir+'/bad_inst'
BADDATE=test_dir+'/bad_date'
BADWAVE=test_dir+'/bad_wave'
BADMOD=test_dir+'/bad_mod'
BADSHAPE=test_dir+'/bad_shape'
DLDATA=test_dir+'/DL'
CRYODATA=test_dir+'/cryo'
LOWFLUX=test_dir+'/lowflux'
if (not os.path.exists(RUN1) and not os.path.isdir(RUN1))\
        or (not os.path.exists(RUN2) and not os.path.isdir(RUN2))\
        or (not os.path.exists(BADINST) and not os.path.isdir(BADINST))\
        or (not os.path.exists(BADDATE) and not os.path.isdir(BADDATE))\
        or (not os.path.exists(BADWAVE) and not os.path.isdir(BADWAVE))\
        or (not os.path.exists(BADMOD) and not os.path.isdir(BADMOD))\
        or (not os.path.exists(BADSHAPE) and not os.path.isdir(BADSHAPE))\
        or (not os.path.exists(DLDATA) and not os.path.isdir(DLDATA))\
        or (not os.path.exists(LOWFLUX) and not os.path.isdir(LOWFLUX))\
        or (not os.path.exists(CRYODATA) and not os.path.isdir(CRYODATA)):
    gen_fake_data.make_testing_data()

def get_telescope_geometry(PCD):

    times = PCD.timeobs

    ele = np.zeros(PCD.numsteps)
    az = np.zeros(PCD.numsteps)
    tab = np.zeros(PCD.numsteps)
    for n in range(PCD.numsteps):
        obstime = Time(times[n], format='mjd')
        tele, taz, ttab = gen_fake_data.compute_telgeom(obstime)
        ele[n] = tele
        az[n] = taz
        tab[n] = ttab

    return az, ele, tab

class TestMuellerMatrices(TestCase):

    def test_linear_retarder(self):
        np.testing.assert_allclose(Data.linear_retarder(1,0,0), np.diag(np.ones(4)), rtol=RTOL)

    def test_rotation(self):
        np.testing.assert_allclose(Data.rotation(0), np.diag(np.ones(4)), rtol=RTOL)

    def test_mirror(self):
        np.testing.assert_allclose(Data.mirror(1,0), np.diag(np.ones(4)), rtol=RTOL)

class TestInit(TestCase):

    def test_bare_init(self):
        PCD = Data.Drawer()

        self.assertEqual(PCD.data_list, [])
        self.assertEqual(PCD.instrument, '')
        self.assertEqual(PCD.theta_ret_steps.size, 0)

    def test_with_instrument(self):
        PCD = Data.Drawer(instrument='visp')

        self.assertEqual(PCD.modid_key, 'VISP_011')
        self.assertEqual(PCD.modnum_key, 'VISP_010')

    def test_with_directory(self):
        PCD = Data.Drawer(data_dir=RUN1)

        self.assertEqual(PCD.instrument.lower(), 'visp')
        self.assertEqual(PCD.theta_ret_steps.size, 5)

class TestLoadFromDir(TestCase):

    def test_basic_load(self):
        PCD = Data.Drawer()
        PCD.load_from_dir(RUN1)

        true_az, true_ele, true_tab = get_telescope_geometry(PCD)

        self.assertEqual(PCD.instrument.lower(), 'visp')
        self.assertEqual(PCD.nummod, 10)
        self.assertEqual(PCD.numsteps, 5)
        self.assertEqual(PCD.wavelength, 620.)
        self.assertEqual(PCD.date_bgn, 57530)
        self.assertAlmostEqual(PCD.date_end, 57530.00104166667)
        self.assertEqual(len(PCD.data_list), 5)
        self.assertEqual(len(PCD.data_list[0]), 11)
        np.testing.assert_array_equal(PCD.theta_pol_steps, np.linspace(0, 360, 5))
        np.testing.assert_array_equal(PCD.theta_ret_steps, np.array([27.5, 22.5, 27.5, 22.5, 27.5]))
        np.testing.assert_array_equal(PCD.pol_in, np.array([True, True, True, True, True]))
        np.testing.assert_array_equal(PCD.ret_in, np.array([True, False, True, True, False]))
        np.testing.assert_array_almost_equal(PCD.timeobs, np.arange(5) * 15 / (3600 * 24) + 57530 + 15 / (3600 * 24))
        np.testing.assert_allclose(PCD.azimuth, true_az)
        np.testing.assert_allclose(PCD.elevation, true_ele)
        np.testing.assert_allclose(PCD.table_angle, true_tab)

    def test_no_skip_darks(self):
        PCD = Data.Drawer()
        PCD.load_from_dir(RUN1, skip_darks=False)

        true_az, true_ele, true_tab = get_telescope_geometry(PCD)

        self.assertEqual(PCD.instrument.lower(), 'visp')
        self.assertEqual(PCD.nummod, 10)
        self.assertEqual(PCD.numsteps, 7)
        self.assertEqual(PCD.wavelength, 620.)
        self.assertEqual(PCD.date_bgn, 57530)
        self.assertAlmostEqual(PCD.date_end, 57530.00104166667)
        self.assertEqual(len(PCD.data_list), 7)
        self.assertEqual(len(PCD.data_list[0]), 11)
        np.testing.assert_array_equal(PCD.theta_pol_steps, np.r_[np.array([0]), np.linspace(0, 360, 5), np.array([0])])
        np.testing.assert_array_equal(PCD.theta_ret_steps, np.array([22.5, 27.5, 22.5, 27.5, 22.5, 27.5, 22.5]))
        np.testing.assert_array_equal(PCD.pol_in, np.array([True, True, True, True, True, True, True]))
        np.testing.assert_array_equal(PCD.ret_in, np.array([True, True, False, True, True, False, True]))
        np.testing.assert_array_almost_equal(PCD.timeobs, np.arange(7) * 15 / (3600 * 24) + 57530)
        np.testing.assert_allclose(PCD.azimuth, true_az)
        np.testing.assert_allclose(PCD.elevation, true_ele)
        np.testing.assert_allclose(PCD.table_angle, true_tab)

    def test_bad_instrument(self):
        PCD = Data.Drawer()
        with self.assertRaisesRegex(ValueError, 'Not all input files were taken with the same instrument'):
            PCD.load_from_dir(BADINST)

    @pytest.mark.skip
    def test_bad_date(self):
        PCD = Data.Drawer()
        with self.assertRaisesRegex(ValueError, 'Not all input files have the same start/end dates'):
            PCD.load_from_dir(BADDATE)

    def test_bad_wavelength(self):
        PCD = Data.Drawer()
        with self.assertRaisesRegex(ValueError, 'Not all input files were taken at the same wavelength'):
            PCD.load_from_dir(BADWAVE)

    def test_bad_modnum(self):
        PCD = Data.Drawer()
        with self.assertRaisesRegex(ValueError, 'Not all input files have the same number of modulator states'):
            PCD.load_from_dir(BADMOD)

class TestIntensityFit(TestCase):

    def test_low_clear_flux(self):

        PCD = Data.Drawer(LOWFLUX, remove_I_trend=False)
        with self.assertRaisesRegex(ValueError, 'Flux in Clear measurements is too low'):
            PCD.find_clears()
            PCD.fit_intensity_trend()

class TestAddition(TestCase):

    def test_concatenation(self):
        PCD1 = Data.Drawer(RUN1)
        PCD2 = Data.Drawer(RUN2)
        PCD = PCD1 + PCD2

        az1, ele1, tab1 = get_telescope_geometry(PCD1)
        az2, ele2, tab2 = get_telescope_geometry(PCD2)

        true_az = np.r_[az1, az2]
        true_ele = np.r_[ele1, ele2]
        true_tab = np.r_[tab1, tab2]

        self.assertEqual(PCD.instrument.lower(), 'visp')
        self.assertEqual(PCD.nummod, 10)
        self.assertEqual(PCD.numsteps, 10)
        self.assertEqual(PCD.wavelength, 620.)
        self.assertEqual(PCD.date_bgn, 57530)
        self.assertEqual(PCD.RN, 13.)
        self.assertAlmostEqual(PCD.date_end, 57560.00104166667)
        self.assertEqual(len(PCD.data_list), 10)
        self.assertEqual(len(PCD.data_list[0]), 11)
        np.testing.assert_array_equal(PCD.theta_pol_steps, np.r_[np.linspace(0, 360, 5), np.ones(5)*180])
        np.testing.assert_array_equal(PCD.theta_ret_steps, np.r_[np.array([27.5, 22.5, 27.5, 22.5, 27.5]), np.linspace(0,360,5)])
        np.testing.assert_array_equal(PCD.pol_in, np.array([True, True, True, True, True, True, True, True, True, True]))
        np.testing.assert_array_equal(PCD.ret_in, np.array([True, False, True, True, False, True, False, True, True, False]))
        np.testing.assert_array_almost_equal(PCD.timeobs, 15 / (24 * 3600) + np.r_[np.arange(5) * 15 / (3600 * 24) + 57530, np.arange(5) * 15/(3600*24) + 57560.])
        np.testing.assert_allclose(PCD.azimuth, true_az)
        np.testing.assert_allclose(PCD.elevation, true_ele)
        np.testing.assert_allclose(PCD.table_angle, true_tab)

    def test_new_object(self):
        PCD1 = Data.Drawer(RUN1)
        PCD2 = Data.Drawer(RUN2)
        PCD = PCD1 + PCD2

        self.assertIsNot(PCD1, PCD)

class TestSlice(TestCase):

    def setUp(self):
        self.PCD = Data.Drawer(RUN1)

    def tearDown(self):
        del self.PCD

    def test_basic_slice(self):
        I = self.PCD[2,1,0]
        self.assertEqual(I.shape, (10,5))
        self.assertIs(type(I), np.ndarray)
        self.assertEqual(I.dtype, np.float64)

    def test_bad_slice(self):
        with self.assertRaisesRegex(IndexError, 'Drawer must be indexed by exactly three values'):
            I = self.PCD[1,0]

    def test_bad_slice2(self):
        with self.assertRaisesRegex(IndexError, 'Drawer must be indexed by exactly three values'):
            I = self.PCD[0]

    def test_non_int(self):
        with self.assertRaisesRegex(IndexError, 'Only integers are allowed as valid indices'):
            I = self.PCD[0,1.1,0]

    def test_colon(self):
        with self.assertRaisesRegex(IndexError, 'Only integers are allowed as valid indices'):
            I = self.PCD[0,:,0]

class TestShapes(TestCase):

    def test_shape(self):
        PCD = Data.Drawer(RUN1)
        self.assertEqual(PCD.shape, (3,2,1))

    def test_wrong_shape(self):
        with self.assertRaisesRegexp(ValueError,'Data do not appear to have 3 dimensions.*'):
            PCD = Data.Drawer(BADSHAPE)

class TestUncertainty(TestCase):
    def setUp(self):
        self.PCD = Data.Drawer(RUN2)

    def tearDown(self):
        del self.PCD

    def test_correct_RN(self):
        self.assertEqual(self.PCD.RN, 13.)

    def test_correct_unc(self):
        I = self.PCD[0,0,0]
        u = self.PCD.get_uncertainty(I)
        np.testing.assert_array_equal(u, np.sqrt(np.abs(I) + self.PCD.RN**2))

    def test_correct_dresser(self):
        PCD1 = Data.Drawer(RUN1)
        I1 = PCD1[0,0,0]
        I2 = self.PCD[0,0,0]
        DRSR = Data.Dresser()
        DRSR.add_drawer(PCD1)
        DRSR.add_drawer(self.PCD)
        I, u = DRSR[0,0,0]
        np.testing.assert_array_equal(u, np.hstack((PCD1.get_uncertainty(I1), self.PCD.get_uncertainty(I2))))

class TestDresser(TestCase):

    def setUp(self):
        self.DRSR = Data.Dresser()
        self.PCD1 = Data.Drawer(RUN1)
        self.DRSR.add_drawer(self.PCD1)
        self.PCD2 = Data.Drawer(RUN2)
        self.DRSR.add_drawer(self.PCD2)

    def tearDown(self):
        del self.DRSR
        del self.PCD1
        del self.PCD2

    def test_numdrawer(self):
        self.assertEqual(self.DRSR.numdrawers, 2)

    def test_numsteps(self):
        self.assertEqual(self.DRSR.numsteps, 10)

    def test_numsteps_nodarkskip(self):
        DRSR = Data.Dresser()
        DRSR.add_drawer(Data.Drawer(RUN1, skip_darks=False))
        DRSR.add_drawer(Data.Drawer(RUN2, skip_darks=False))
        self.assertEqual(DRSR.numsteps, 14)

    def test_step_list(self):
        self.assertEqual(self.DRSR.drawer_step_list, [5,5])

    def test_step_list_nodarkskip(self):
        DRSR = Data.Dresser()
        DRSR.add_drawer(Data.Drawer(RUN1, skip_darks=False))
        DRSR.add_drawer(Data.Drawer(RUN2, skip_darks=False))
        self.assertEqual(DRSR.drawer_step_list, [7,7])

    def test_shape(self):
        self.assertEqual(self.DRSR.shape, (3,2,1))

    def test_wrong_modnum(self):
        DRSR = Data.Dresser()
        DRSR.nummod = 99
        with self.assertRaisesRegex(ValueError, 'Trying to add Drawer with 10 mod states to Dresser with 99'):
            DRSR.add_drawer(Data.Drawer(RUN1))

    def test_wrong_wave(self):
        DRSR = Data.Dresser()
        DRSR.wavelength = 999
        with self.assertRaisesRegex(ValueError, 'Drawer with wave = .*620\.0 cannot be added to Dresser with wave = .*999\.0'):
            DRSR.add_drawer(Data.Drawer(RUN1))

    def test_wrong_instrument(self):
        DRSR = Data.Dresser()
        DRSR.instrument = 'NOTHING'
        with self.assertRaisesRegex(ValueError, 'Drawer from instrument VISP cannot be added to Dresser from instrument NOTHING'):
            DRSR.add_drawer(Data.Drawer(RUN1))

    def test_wrong_shape(self):
        DRSR = Data.Dresser()
        DRSR.shape = (99,)
        with self.assertRaisesRegex(ValueError, 'Drawer with shape \(3, 2, 1\) does not fit into Dresser with shape \(99,\)'):
            DRSR.add_drawer(Data.Drawer(RUN1))

    def test_get_item_correct(self):
        truth = np.hstack((self.PCD1[0,0,0], self.PCD2[0,0,0]))
        np.testing.assert_array_equal(self.DRSR[0,0,0][0], truth)

class TestVISPKeywords(TestCase):
    def setUp(self):
        self.PCD = Data.Drawer(RUN1)

    def tearDown(self):
        del self.PCD

    def test_modid(self):
        self.assertEqual(self.PCD.data_list[0][1].header[self.PCD.modid_key], 1)
        self.assertEqual(self.PCD.data_list[0][2].header[self.PCD.modid_key], 2)
        self.assertEqual(self.PCD.data_list[0][3].header[self.PCD.modid_key], 3)
        self.assertEqual(self.PCD.data_list[0][4].header[self.PCD.modid_key], 4)
        self.assertEqual(self.PCD.data_list[0][5].header[self.PCD.modid_key], 5)
        self.assertEqual(self.PCD.data_list[0][6].header[self.PCD.modid_key], 6)
        self.assertEqual(self.PCD.data_list[0][7].header[self.PCD.modid_key], 7)
        self.assertEqual(self.PCD.data_list[0][8].header[self.PCD.modid_key], 8)
        self.assertEqual(self.PCD.data_list[0][9].header[self.PCD.modid_key], 9)
        self.assertEqual(self.PCD.data_list[0][10].header[self.PCD.modid_key], 10)

    def test_modnum(self):
        self.assertEqual(self.PCD.data_list[0][0].header[self.PCD.modnum_key], 10)
        self.assertEqual(self.PCD.data_list[0][5].header[self.PCD.modnum_key], 10)


class TestDLNIRSPKeywords(TestCase):
    def setUp(self):
        self.PCD = Data.Drawer(DLDATA)

    def tearDown(self):
        del self.PCD

    def test_modid(self):
        self.assertEqual(self.PCD.data_list[0][1].header[self.PCD.modid_key], 0)
        self.assertEqual(self.PCD.data_list[0][2].header[self.PCD.modid_key], 1)
        self.assertEqual(self.PCD.data_list[0][3].header[self.PCD.modid_key], 2)
        self.assertEqual(self.PCD.data_list[0][4].header[self.PCD.modid_key], 3)
        self.assertEqual(self.PCD.data_list[0][5].header[self.PCD.modid_key], 4)
        self.assertEqual(self.PCD.data_list[0][6].header[self.PCD.modid_key], 5)
        self.assertEqual(self.PCD.data_list[0][7].header[self.PCD.modid_key], 6)
        self.assertEqual(self.PCD.data_list[0][8].header[self.PCD.modid_key], 7)
        self.assertEqual(self.PCD.data_list[0][9].header[self.PCD.modid_key], 8)
        self.assertEqual(self.PCD.data_list[0][10].header[self.PCD.modid_key], 9)

    def test_modnum(self):
        self.assertEqual(self.PCD.data_list[0][0].header[self.PCD.modnum_key], 10)
        self.assertEqual(self.PCD.data_list[0][5].header[self.PCD.modnum_key], 10)

class TestCRYONIRSPKeywords(TestCase):
    def setUp(self):
        self.PCD = Data.Drawer(CRYODATA)

    def tearDown(self):
        del self.PCD

    def test_modid(self):
        self.assertEqual(self.PCD.data_list[0][1].header[self.PCD.modid_key], 0)
        self.assertEqual(self.PCD.data_list[0][2].header[self.PCD.modid_key], 1)
        self.assertEqual(self.PCD.data_list[0][3].header[self.PCD.modid_key], 2)
        self.assertEqual(self.PCD.data_list[0][4].header[self.PCD.modid_key], 3)
        self.assertEqual(self.PCD.data_list[0][5].header[self.PCD.modid_key], 4)
        self.assertEqual(self.PCD.data_list[0][6].header[self.PCD.modid_key], 5)
        self.assertEqual(self.PCD.data_list[0][7].header[self.PCD.modid_key], 6)
        self.assertEqual(self.PCD.data_list[0][8].header[self.PCD.modid_key], 7)
        self.assertEqual(self.PCD.data_list[0][9].header[self.PCD.modid_key], 8)
        self.assertEqual(self.PCD.data_list[0][10].header[self.PCD.modid_key], 9)

    def test_modnum(self):
        self.assertEqual(self.PCD.data_list[0][0].header[self.PCD.modnum_key], 10)
        self.assertEqual(self.PCD.data_list[0][5].header[self.PCD.modnum_key], 10)
